#!/bin/bash

################################################################################
#   Author : Yves Vanaubel                                                     #
#   Date : 08/01/2015                                                          #
#                                                                              #
################################################################################

shopt -s extglob

if [ $# -ne 3 ]
then
    echo 'Usage : ./apply_LPRA.sh <ts_file> <output_dir> <filtering_type>'
    echo 'Filtering type is 1 if strict AS diversity is applied on static filtering'
    exit 1
fi

script_dir=$(dirname $0)
root_script_dir=$script_dir/../

ts_file=$1
output_dir=$2
filtering_type=$3

# Generating output directories
if [ -e $output_dir ]
then
    rm -rf $output_dir
fi
mkdir $output_dir

mkdir $output_dir/TE_MAIN_STATIC_FILTERED

mkdir $output_dir/MPLS_ARRAY
mkdir $output_dir/MPLS_ARRAY_STATIC_FILTERED

mkdir $output_dir/TOP_ASES

mkdir $output_dir/TOP_ASES_TEMPORAL_FILTERED

mkdir $output_dir/STATS

mkdir $output_dir/CLASSIFICATION
mkdir $output_dir/CLASSIFICATION/MODE1
mkdir $output_dir/CLASSIFICATION/MODE2_EXPANDED
mkdir $output_dir/CLASSIFICATION_NO_TEMP_FILTER
mkdir $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE1
mkdir $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE2_EXPANDED

# Get main cycle
group_ts_file=$output_dir/group_ts.txt
awk '{print $1}' $ts_file > $group_ts_file
main_cycle_ts=$(sort -n -k 1.7,1.10 -k 1.4,1.5 $group_ts_file | head -n 1)

# Get the cycle, the corresponding month and year
main_cycle_year=$(echo $main_cycle_ts | awk '{split($0, d, "-"); print d[3]}')
main_cycle_month=$(echo $main_cycle_ts | awk '{split($0, d, "-"); print d[2]}')
main_cycle_prefix_dir=prefix2as/$main_cycle_year/$main_cycle_month/$main_cycle_ts
main_cycle_tunnels_dir=tunnels/$main_cycle_year/$main_cycle_month/$main_cycle_ts

# Decompress prefix to AS file
rm_prefixes=0
unzip_main_prefix2as_file=$main_cycle_prefix_dir/prefix2as.txt
if [ ! -e $unzip_main_prefix2as_file ]
then
    gunzip -c $main_cycle_prefix_dir/prefix2as.txt.gz > $unzip_main_prefix2as_file
    rm_prefixes=1
fi

# Decompress tunnels file
rm_tunnels=0
unzip_main_tunnels_file=$main_cycle_tunnels_dir/etun_dst_all.txt
unzip_main_ips_file=$main_cycle_tunnels_dir/all.ips
if [ ! -e $unzip_main_tunnels_file ]
then
    gunzip -c $main_cycle_tunnels_dir/etun_dst_all.txt.gz > $unzip_main_tunnels_file
    gunzip -c $main_cycle_tunnels_dir/all.ips.gz > $unzip_main_ips_file
    rm_tunnels=1
fi

# Get AS MPLS array before filtering
python $script_dir/get_asn_analyze.py $output_dir/MPLS_ARRAY $main_cycle_tunnels_dir $unzip_main_prefix2as_file

# Apply static filtering
echo "Starting static filtering..."
python $script_dir/apply_static_filters.py $output_dir/TE_MAIN_STATIC_FILTERED $main_cycle_tunnels_dir $filtering_type

awk '{$1 = ""; gsub(/^ /, "", $0); print $0}' $output_dir/TE_MAIN_STATIC_FILTERED/filtered_tunnels.txt > $output_dir/TE_MAIN_STATIC_FILTERED/etun_dst_all.txt
cp $main_cycle_tunnels_dir/all.ips $output_dir/TE_MAIN_STATIC_FILTERED/

# Get AS MPLS array after filtering
python $script_dir/get_asn_analyze.py $output_dir/MPLS_ARRAY_STATIC_FILTERED $output_dir/TE_MAIN_STATIC_FILTERED $unzip_main_prefix2as_file

if [ $rm_prefixes -eq 1 ]
then
    rm $unzip_main_prefix2as_file
fi
if [ $rm_tunnels -eq 1 ]
then
    rm $unzip_main_ips_file
fi

echo "Selecting top ASes..."
# Select TOP Ases
python $script_dir/select_top_ASes.py $output_dir/TOP_ASES $output_dir/TE_MAIN_STATIC_FILTERED/filtered_tunnels.txt $output_dir/MPLS_ARRAY_STATIC_FILTERED/as_array.txt 0

echo "Processing stats..."

# Get some stats
cur_dir=$output_dir/TE_MAIN_STATIC_FILTERED/

nb_tunnels=$(echo $(wc -l < $main_cycle_tunnels_dir/etun_dst_all.txt))
nb_VPNs=$(echo $(grep '|' $main_cycle_tunnels_dir/etun_dst_all.txt | wc -l))
initial_prop_VPNs=$(echo "$nb_VPNs $nb_tunnels" | awk '{printf "%.3f", $1 / $2}' | sed -e 's/,/./g')

if [ $rm_tunnels -eq 1 ]
then
    rm $unzip_main_tunnels_file
fi

nb_NA_AS=$(echo $(wc -l < $cur_dir/NA_AS.txt))
nb_dst_one_AS=$(echo $(wc -l < $cur_dir/dst_one_as.txt))
nb_lsp_star=$(echo $(wc -l < $cur_dir/lsp_star.txt))
nb_dst_same_AS=$(echo $(wc -l < $cur_dir/dst_same_as.txt))
nb_multiple_AS=$(echo $(wc -l < $cur_dir/multiple_as.txt))
nb_dst_one_as_multi_LSPs=$(echo $(wc -l < $cur_dir/dst_one_as_multi_LSPs.txt))
nb_ie_star=$(echo $(wc -l < $cur_dir/ie_star.txt))
nb_tunnels_after_static_filter=$(echo $(wc -l < $cur_dir/filtered_tunnels.txt))
let 'nb_deleted_tunnels_static_filter = nb_tunnels - nb_tunnels_after_static_filter'
let 'nb_tunnels_completely_multiple_ASes = nb_tunnels - nb_dst_one_AS - nb_dst_same_AS - nb_multiple_AS - nb_ie_star - nb_lsp_star - nb_tunnels_after_static_filter - nb_NA_AS'

# Compressing data not useful anymore
rm $output_dir/TE_MAIN_STATIC_FILTERED/filtered_tunnels.txt $output_dir/TE_MAIN_STATIC_FILTERED/all.ips
cd $cur_dir
gzip *
cd - > /dev/null 2>&1

echo "Starting temporal filtering..."
## Apply temporal filter
python $script_dir/apply_temporal_filter.py $output_dir/TOP_ASES_TEMPORAL_FILTERED $output_dir/TOP_ASES tunnels $group_ts_file $output_dir/STATS 0

echo "Updating stats after temporal filtering..."

nb_tunnels_top_ases=$(echo $(awk 'BEGIN{c=0}{c+=$5}END{print c}' $output_dir/STATS/temporal_stats.txt))
nb_compliant_tunnels=$(echo $(awk 'BEGIN{c=0}{c+=$3}END{print c}' $output_dir/STATS/temporal_stats.txt))
nb_compliant_VPNs=$(echo $(awk 'BEGIN{c=0}{c+=$2}END{print c}' $output_dir/STATS/temporal_stats.txt))
nb_side_effect_tunnels=$(echo $(awk 'BEGIN{c=0}{c+=$4}END{print c}' $output_dir/STATS/temporal_stats.txt))
nb_top_ases=$(echo $(wc -l < $output_dir/TOP_ASES/top_ases.txt))

let 'nb_tunnels_other_ases = nb_tunnels_after_static_filter - nb_tunnels_top_ases'
let 'nb_deleted_tunnels_temporal_filter = nb_tunnels_top_ases - nb_compliant_tunnels'
prop_VPNs_after_temporal_filter=$(echo "$nb_compliant_VPNs $nb_compliant_tunnels" | awk '{printf "%.3f", $1 / $2}' | sed -e 's/,/./g')

echo "Number of tunnels for this cycle : $nb_tunnels" > $output_dir/STATS/stats.txt
echo "Number of VPNs : $nb_VPNs" >> $output_dir/STATS/stats.txt
echo "=> Proportion of VPNs : $initial_prop_VPNs" >> $output_dir/STATS/stats.txt
echo >> $output_dir/STATS/stats.txt
echo >> $output_dir/STATS/stats.txt
echo "Number of tunnels with ingress or egress being * : $nb_ie_star" >> $output_dir/STATS/stats.txt
echo "Number of tunnels with at least one * in LSP : $nb_lsp_star" >> $output_dir/STATS/stats.txt
echo "Number of tunnels contaning at least one NA IP address : $nb_NA_AS" >> $output_dir/STATS/stats.txt
echo >> $output_dir/STATS/stats.txt
echo "Number of tunnels with only 1 destination AS : $nb_dst_one_AS" >> $output_dir/STATS/stats.txt
echo "Number of tunnels in same AS as the destination : $nb_dst_same_AS" >> $output_dir/STATS/stats.txt
echo "Number of tunnels in multiple ASes (hops in different ASes) : $nb_multiple_AS" >> $output_dir/STATS/stats.txt
echo "Number of tunnels that appears completely in multiple ASes (hops in same AS) : $nb_tunnels_completely_multiple_ASes" >> $output_dir/STATS/stats.txt
echo >> $output_dir/STATS/stats.txt
echo "Number of tunnels after static filtering : $nb_tunnels_after_static_filter" >> $output_dir/STATS/stats.txt
echo "=> $nb_deleted_tunnels_static_filter tunnels were deleted by static filtering" >> $output_dir/STATS/stats.txt
echo >> $output_dir/STATS/stats.txt
echo "Top $nb_top_ases AS : tunnels : $nb_tunnels_top_ases" >> $output_dir/STATS/stats.txt
echo "=> $nb_tunnels_other_ases tunnels appear in other ASes" >> $output_dir/STATS/stats.txt
echo >> $output_dir/STATS/stats.txt
echo "After temporal filtering : $nb_compliant_tunnels" >> $output_dir/STATS/stats.txt
echo "=> $nb_deleted_tunnels_temporal_filter tunnels were deleted by temporal filtering, containing $nb_side_effect_tunnels tunnels deleted by side effect" >> $output_dir/STATS/stats.txt
echo "After temporal filtering : $nb_compliant_VPNs VPNs" >> $output_dir/STATS/stats.txt
echo "=> Proportion of VPNs : $prop_VPNs_after_temporal_filter" >> $output_dir/STATS/stats.txt
echo >> $output_dir/STATS/stats.txt
echo "NB: If a tunnel (I -> E) appears for X destinations, it is taken into account X times" >> $output_dir/STATS/stats.txt

echo "Starting classification..."
# Apply classification with temporal filter
python $script_dir/apply_classification.py $output_dir/CLASSIFICATION/MODE1/classes $output_dir/TOP_ASES_TEMPORAL_FILTERED 0
python $script_dir/apply_classification.py $output_dir/CLASSIFICATION/MODE2_EXPANDED/classes $output_dir/TOP_ASES_TEMPORAL_FILTERED 1

# Compress the results
cd $output_dir/TOP_ASES_TEMPORAL_FILTERED
gzip *
cd - > /dev/null 2>&1

# Get the distributions
$script_dir/get_LPRA_distributions.sh $output_dir/CLASSIFICATION/MODE1/classes $output_dir/CLASSIFICATION/MODE1/general_distribution 0
$script_dir/get_LPRA_distributions.sh $output_dir/CLASSIFICATION/MODE1/classes $output_dir/CLASSIFICATION/MODE1/detail_distribution 1
$script_dir/get_LPRA_distributions.sh $output_dir/CLASSIFICATION/MODE2_EXPANDED/classes $output_dir/CLASSIFICATION/MODE2_EXPANDED/general_distribution 0
$script_dir/get_LPRA_distributions.sh $output_dir/CLASSIFICATION/MODE2_EXPANDED/classes $output_dir/CLASSIFICATION/MODE2_EXPANDED/detail_distribution 1


# Apply classification with no temporal filter
python $script_dir/apply_classification.py $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE1/classes $output_dir/TOP_ASES 0
python $script_dir/apply_classification.py $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE2_EXPANDED/classes $output_dir/TOP_ASES 1

# Compress the results
cd $output_dir/TOP_ASES
gzip *
cd - > /dev/null 2>&1

# Get the distributions
$script_dir/get_LPRA_distributions.sh $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE1/classes $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE1/general_distribution 0
$script_dir/get_LPRA_distributions.sh $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE1/classes $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE1/detail_distribution 1
$script_dir/get_LPRA_distributions.sh $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE2_EXPANDED/classes $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE2_EXPANDED/general_distribution 0
$script_dir/get_LPRA_distributions.sh $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE2_EXPANDED/classes $output_dir/CLASSIFICATION_NO_TEMP_FILTER/MODE2_EXPANDED/detail_distribution 1

exit 0
