#!/usr/bin/env python

################################################################################
# apply_LPRA.py                                                                #
#                                                                              #
# Author Yves Vanaubel (yves.vanaubel@ulg.ac.be)                               #
# Date 09/01/15                                                                #
# Description: Apply the Label Pattern Recognition Algorithm (LPR)             #
#              See paper for complete description                              #
# extended_commmon_ips = 1 if must extend common IP set with the previous IPs  #
################################################################################

# IMPORTS
import sys
import re
import os
import numpy as np
import subprocess
from collections import defaultdict
from optparse import OptionParser

def main():
	usage = "usage: %prog <output_dir> <top_as_dir> <extend_common_IPs>"
	
	parser = OptionParser(usage=usage)
	
	(options, args) = parser.parse_args()
    
	#Check arg number
	if not len(args) == 3:
	   parser.print_help()
	   sys.exit(1)
	run(args)

################################################################################

def run(args):
    
    output_dir = args[0]
    top_as_dir = args[1]
    extend_common_IPs = int(args[2])
    
    # Create the output directory
    os.system('rm -rf ' + output_dir)
    os.system('mkdir ' + output_dir)

    # Get the different AS files
    as_files_string = subprocess.check_output('ls ' + top_as_dir + ' | grep -v top_ases.txt', shell=True)
    as_files = as_files_string.split()

    for as_file in as_files:
        # Different tunnels and destinations
        tunnels_dsts = defaultdict(lambda: defaultdict(list))
    
        # Extract the ASn
        asn = re.search('as_(\d+).txt', as_file).group(1)
        
        # Get each tunnel
        tunnels_file = open(top_as_dir + '/' + as_file)
        for tunnel in tunnels_file:
            tunnel = tunnel.strip()
            # Skip empty lines
            if tunnel == "":
                continue

            # Get the destination and the tunnel
            tun_dst = tunnel.split(" => ")
            # Get the ingress egress pair
            nodes = tun_dst[1].split()
            egress_index = len(nodes) - 1
            ingress_egress = '<' + nodes[0] + ',' + nodes[egress_index] + '>'
            
            # Get the configuration of the tunnel
            config = nodes[1]
            for n in xrange(2, egress_index):
                config += ' ' + nodes[n]

            # Add the tunnel
            tunnels_dsts[ingress_egress]["tun"].append(config)
            tunnels_dsts[ingress_egress]["dst"].append(tun_dst[2])
    
        tunnels_file.close()


        # Start classification
        ecmp = False
        ecmp_aliases = False
        output_file_ECMP = ""
        output_file_ECMP_multi_aliases = ""
        output_file_classes = open(output_dir + '/classes_' + asn + '.txt' , 'w')
        output_file_classes.write('{0:<50} {1:<16} {2:<16} {3:<20} {4:<16} {5:<16} {6:<16}\n'.format('#I-E', '#LPR_class', '#max_LSP_length', '#symetry', '#width', '#width_IP', '#width_lab'))

        # Get each pair INGRESS-EGRESS
        for ingress_egress in tunnels_dsts:
            # Number of destinations and configurations for the tunnel
            nb_dst = len(tunnels_dsts[ingress_egress]["dst"])
            nb_config = np.unique(tunnels_dsts[ingress_egress]["tun"]).size
            
            ie_class = 0
            width = 1
            width_IP = 1
            width_lab = 1
            max_LSP_length = 0
            min_LSP_length = 0
            
            # CLASS 1 : mono LSP, mono FEC
            if nb_config == 1:
                ie_class = '1'
                max_LSP_length = len(tunnels_dsts[ingress_egress]["tun"][0].split())
                min_LSP_length = max_LSP_length
            # Need a deeper analyze
            else:
                # Collect data
                labels = list()
                ips = list()
                ips_split = list()
                ip2label = defaultdict(list)
                fh_ips = list()
                ip2previous_ips = defaultdict(set)

                # Get egress
                egress_ip = re.search('<\d+\.\d+\.\d+\.\d+:\w+,(\d+\.\d+\.\d+\.\d+):\w+>', ingress_egress).group(1)
                for i in xrange(nb_dst):
                    # Replace with "-" in case of no label
                    c = re.sub("(\d+\.\d+\.\d+\.\d+):(\w+)(?:\s+|$)",r'\1:\2:- ', tunnels_dsts[ingress_egress]["tun"][i]).strip()
                    cur_ips = re.sub("(\d+\.\d+\.\d+\.\d+):\w+:*([\d|-]*)", r'\1', c)
                    cur_labels = re.sub("(\d+\.\d+\.\d+\.\d+):\w+:*([\d|-]*)", r'\2', c)
                    cur_ips_array = cur_ips.split()
                    cur_labels_array = cur_labels.split()
                    ips.append(cur_ips)
                    ips_split.append(cur_ips_array)
                    labels.append(cur_labels)
                    
                    # Length of LSP
                    cur_lsp_length = len(cur_ips_array)
                    if cur_lsp_length > max_LSP_length:
                        max_LSP_length = cur_lsp_length
                    if min_LSP_length == 0:
                        min_LSP_length = cur_lsp_length
                    else:
                        if cur_lsp_length < min_LSP_length:
                            min_LSP_length = cur_lsp_length
                    
                    # Associate the labels to each IP
                    nb_ips = cur_lsp_length
                    for l in range(nb_ips):
                        ip2label[cur_ips_array[l]].append(cur_labels_array[l])
                        # Get the previous labels
                        if l > 0:
                            ip2previous_ips[cur_ips_array[l]].add(cur_ips_array[l - 1])
                        else:
                            fh_ips.append(cur_ips_array[l])
                    
                    # Add egress
                    ip2previous_ips[egress_ip].add(cur_ips_array[-1])
                        
                # Number of configurations of LSP (IPs and Labels)
                nb_config_stacks = np.unique(labels).size
                nb_config_ips = np.unique(ips).size
                width = nb_config
                width_IP = nb_config_ips
                width_lab = nb_config_stacks

                # Analyze the data
                m, n, k, p = 0, 0, 0, 0
                nb_common_ips = 0
                
                # Get the common IPs
                common_ips_set = set()
                for ip in ip2label:
                    if len(ip2label[ip]) > 1:
                        common_ips_set.add(ip)
                # Add egress
                common_ips_set.add(egress_ip)
                
                # Extended set
                if extend_common_IPs == 1:
                    common_ips_set_copy = common_ips_set.copy()
                    for ip in common_ips_set_copy:
                        if (ip not in fh_ips) and (ip in ip2previous_ips):
                            previous_ips = ip2previous_ips[ip].copy()
                            inters_common_ips = previous_ips.intersection(common_ips_set)
                            prev_ip = ''
                            if inters_common_ips:
                                prev_ip = inters_common_ips.pop()
                                previous_ips.remove(prev_ip)
                                for inter_ip in inters_common_ips:
                                    for inter_lab in ip2label[inter_ip]:
                                        ip2label[prev_ip].append(inter_lab)
                                    previous_ips.remove(inter_ip)
                                    common_ips_set.remove(inter_ip)
                            else:
                                prev_ip = previous_ips.pop()
                                common_ips_set.add(prev_ip)
                            
                            for other_ip in previous_ips:
                                for other_lab in ip2label[other_ip]:
                                    ip2label[prev_ip].append(other_lab)
                        
                nb_common_ips = len(common_ips_set)
                
                # Analyze the common IPs
                for ip in common_ips_set:
                    nb_diff_labels = np.unique(ip2label[ip]).size
                    p += nb_diff_labels
                    
                    # Get the number of different labels for previous IPs, if they exist
                    prev_labels = list()
                    for previous_ip in ip2previous_ips[ip]:
                        for lab in ip2label[previous_ip]:
                            prev_labels.append(lab)
                    nb_diff_previous_labels = np.unique(prev_labels).size
                    
                    if nb_diff_labels > 1:
                        n += nb_diff_labels
                        if (ip not in fh_ips) and (nb_diff_previous_labels > 0) and (nb_diff_labels > nb_diff_previous_labels):
                            k = 1
                    elif (nb_diff_labels == 1) or (ip not in fh_ips and nb_diff_previous_labels == 1):
                        m += 1
                
                # Multi FEC
                if n  > 0:
                    r = 0
                    if nb_config_ips == 1:
                        r = 1 + (n / float(p))
                    else:
                        r = n / float(p)

                    # CLASS 2 : multi IP LSP, no LB
                    if r == 1:
                        ie_class = '2'
                    # CLASS 2.1 : mono IP LSP, no LB
                    elif r == 2:
                        #ie_class = '2.1'
                        ie_class = '2'
                    # CLASS 2.2 : multi IP LSP and LB
                    elif r < 1 and k == 0:
                        #ie_class = '2.2'
                        ie_class = '2.1'
                    # CLASS 2.3 : inconsistent multi IP LSP
                    elif r < 1 and k == 1:
                        #ie_class = '2.3'
                        ie_class = '2.2'
                    # CLASS 2.4 : inconsistent mono IP LSP
                    else:
                        #ie_class = '2.4'
                        ie_class = '2.2'
                # CLASS 3 : Mono FEC and LB
                elif m == nb_common_ips:
                    # Open files
                    if not ecmp:
                        output_file_ECMP = open(output_dir + '/ECMP_' + asn + '.txt' , 'w')
                        output_file_ECMP.write('{0:<50} {1:<16} {2:<16} {3:<20} {4:<16} {5:<16} {6:<16}\n'.format('#I-E', '#LPR_class', '#max_LSP_length', '#symetry', '#width', '#width_IP', '#width_lab'))
                        ecmp = True
                    # LB multi-routers
                    if nb_config_stacks > 1:
                        ie_class = '3'
                    # LB multi-links
                    else:
                        if not ecmp_aliases:
                            output_file_ECMP_multi_aliases = open(output_dir + '/aliases_multilink_ECMP_' + asn + '.txt' , 'w')
                        ecmp_aliases = True
                        
                        ie_class = '3.1'
                        # Each tunnel in IOTP has the same number of LSRs by definition
                        nb_LSRs = len(ips_split[0])
                        for i in range(nb_LSRs):
                            aliases = set()
                            for lsr in ips_split:
                                aliases.add(lsr[i])
                            if len(aliases) > 1:
                                    output_file_ECMP_multi_aliases.write(ingress_egress + ' ' + str(i + 1) + ' ' + ' '.join(aliases) + '\n')
                    output_file_ECMP.write('{0:<50} {1:<16} {2:<16} {3:<20} {4:<16} {5:<16} {6:<16}\n'.format(ingress_egress, ie_class, str(max_LSP_length), str(max_LSP_length - min_LSP_length), str(width), str(width_IP), str(width_lab)))
                # CLASS 4 : Not enough information to conclude
                else:
                    ie_class = '4'

            # Write the results
            #for tu in tunnels_dsts[ingress_egress]["tun"]:
            #    output_file_classes.write(tu + '\n')
            output_file_classes.write('{0:<50} {1:<16} {2:<16} {3:<20} {4:<16} {5:<16} {6:<16}\n'.format(ingress_egress, ie_class, str(max_LSP_length), str(max_LSP_length - min_LSP_length), str(width), str(width_IP), str(width_lab)))

        output_file_classes.close()
        if ecmp:
            output_file_ECMP.close()
        if ecmp_aliases:
            output_file_ECMP_multi_aliases.close()

if __name__ == '__main__':
	sys.exit(main())
