#!/usr/bin/env python

################################################################################
# apply_static_filters.py                                                      #
#                                                                              #
# Author Yves Vanaubel (yves.vanaubel@ulg.ac.be)                               #
# Date 08/01/15                                                                #
# Description: Filtering for AS analyze of the TE classes for <I,E> couples    #
#              The three main cycles are filtered. The other for temporal      #
#              analyze are not (to not eliminate information)                  #
#              The analyze is realized for 3 cycles.                           #
#              The data directory must contain 3 directories.                  #
#              Each directory contains the cycle to analyze and the two next   #
#              cycles (for temporal analyze).                                  #
################################################################################

# IMPORTS
import sys
import os
import re
import numpy as np
import subprocess
from collections import defaultdict
from optparse import OptionParser

def main():
	usage = "usage: %prog <output_filter_dir> <cycle_tunnels_dir> <strict_dst_AS_diversity>"
	
	parser = OptionParser(usage=usage)
	
	(options, args) = parser.parse_args()
    
	# Check arg number
	if not len(args) == 3:
	   parser.print_help()
	   sys.exit(1)
	run(args)

################################################################################

def run(args):
    
    output_dir = args[0]
    cycle_tunnels_dir = args[1]
    strict_dst_AS_diversity = False
    if args[2] == '1':
        strict_dst_AS_diversity = True
    
    # Check output directory
    os.system('rm -rf ' + output_dir)
    os.system('mkdir ' + output_dir)
    
    # Output file for tunnels in multiple ASes
    output_file_multiple_as = open(output_dir + '/multiple_as.txt', 'w')
    # Output file for tunnels with destination in only one AS
    output_file_dst_one_as = open(output_dir + '/dst_one_as.txt', 'w')
    # Output file for tunnels with destination in only 1 AS but with multiple LSPs
    output_file_dst_one_as_multi_LSPs = open(output_dir + '/dst_one_as_multi_LSPs.txt', 'w')
    # Output file for tunnels with destination in same AS than tunnel
    output_file_dst_same_as = open(output_dir + '/dst_same_as.txt', 'w')
    # Output file for tunnels with ingress or egress equals to *
    output_file_ie_star = open(output_dir + '/ie_star.txt', 'w')
    # Output file for tunnels with an IP in the LSP equal to *
    output_file_LSP_star = open(output_dir + '/lsp_star.txt', 'w')
    # Output file for tunnels with an IP associated to NA
    output_file_NA_AS = open(output_dir + '/NA_AS.txt', 'w')
    # Output file for tunnels that respect the filters
    output_file_filter = open(output_dir + '/filtered_tunnels.txt', 'w')

    # Tunnels encountered in the cycle
    # ie -> asn -> tun    -> tunnel
    #           -> dstasn -> dst_asn
    ie_tunnels = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    # Get each tunnel in the file
    tunnels_file = open(cycle_tunnels_dir + '/etun_dst_all.txt')
    for tunnel in tunnels_file:
        tunnel = tunnel.strip()
        # Skip empty lines
        if tunnel == "":
            continue
        
        # Get the destination and the tunnel
        tun_dst = tunnel.split(" => ")
        tunnel_nodes = tun_dst[1]

        # Get the tunnel (FH->LH) and the IPs
        nodes = tunnel_nodes.split()
        i = 0
        asn_FH_LH = set()
        last_index = len(nodes) - 1
        
        ## FILTER 0 : Delete if ingress or egress unknown
        if nodes[0] == '*' or nodes[last_index] == '*':
            output_file_ie_star.write(tunnel + '\n')
            continue

        ## FILTER 0bis : Delete LSPs containing stars
        if "*" in tunnel_nodes:
            output_file_LSP_star.write(tunnel + '\n')
            continue

        ie = '<' + nodes[0] + ',' + nodes[last_index] + '>'
        for n in nodes:
            # Get ASn between FH and LH
            if i > 0 and i < last_index:
                node_data = n.split(':')
                asn_FH_LH.add(node_data[1])
            i += 1

        ## FILTER 0ter : Delete LSPs containing NA
        if 'NA' in asn_FH_LH:
            output_file_NA_AS.write(tunnel + '\n')
            continue

        # Add the tunnel if possible
        if asn_FH_LH:
            nb_ases = len(asn_FH_LH)
            
            ## FILTER 1 : Tunnel in one AS
            if nb_ases == 1:
                unique_asn = asn_FH_LH.pop()
                dst_asn = tun_dst[2].split(':')[1]
                
                ## FILTER 2 : AS of dst not the same as AS of tunnel
                if dst_asn == unique_asn:
                    output_file_dst_same_as.write(tunnel + '\n')
                else:
                    ie_tunnels[ie][unique_asn][tun_dst[1]].append(tun_dst[0] + " => " + tun_dst[2])
            else:
                output_file_multiple_as.write(tunnel + '\n')

    tunnels_file.close()
    output_file_ie_star.close()
    output_file_LSP_star.close()
    output_file_multiple_as.close()
    output_file_dst_same_as.close()
    output_file_NA_AS.close()

    ## FILTER 3 : Multiple ASes for destinations
    for ie in ie_tunnels:
        ks = ie_tunnels[ie].keys()
        cur_as = ks[0]
        
        # Multiple ASes for the tunnel
        if len(ks) != 1:
            # Take the AS that appears the most.
            max_nb_tun_as = 0
            for test_tun in ie_tunnels[ie][cur_as]:
                max_nb_tun_as += len(ie_tunnels[ie][cur_as][test_tun])
            for i in range(1, len(ks)):
                tmp_nb_tun_as = 0
                for test_tun in ie_tunnels[ie][ks[i]]:
                    tmp_nb_tun_as += len(ie_tunnels[ie][ks[i]][test_tun])
                if tmp_nb_tun_as > max_nb_tun_as:
                    max_nb_tun_as = tmp_nb_tun_as
                    cur_as = ks[i]

        tun_all_dst = set()
        for test_tun in ie_tunnels[ie][cur_as]:
            for hop_dst in ie_tunnels[ie][cur_as][test_tun]:
                tun_all_dst.add(hop_dst.split(" => ")[1].split(':')[1])
        nb_ases_dst = len(tun_all_dst)

        # Multiple ASes for destinations => ACCEPTED
        if nb_ases_dst > 1:
            for t in ie_tunnels[ie][cur_as]:
                for hop_dst in ie_tunnels[ie][cur_as][t]:
                    hop_dst_split = hop_dst.split(" => ")
                    output_file_filter.write(cur_as + ' ' + hop_dst_split[0] + ' => ' + t + ' => ' + hop_dst_split[1] + '\n')
        # Only one AS as destination
        else:
            nb_different_LSPs = len(ie_tunnels[ie][cur_as])
            # At least two different LSPs => ACCEPTED
            if nb_different_LSPs > 1:
                for t in ie_tunnels[ie][cur_as]:
                    for hop_dst in ie_tunnels[ie][cur_as][t]:
                        hop_dst_split = hop_dst.split(" => ")
                        t_string = hop_dst_split[0] + ' => ' + t + ' => ' + hop_dst_split[1]
                        if not strict_dst_AS_diversity:
                            output_file_filter.write(cur_as + ' ' + t_string + '\n')
                        output_file_dst_one_as_multi_LSPs.write(t_string + '\n')
            if strict_dst_AS_diversity or nb_different_LSPs <= 1:
                for t in ie_tunnels[ie][cur_as]:
                    for hop_dst in ie_tunnels[ie][cur_as][t]:
                        hop_dst_split = hop_dst.split(" => ")
                        output_file_dst_one_as.write(hop_dst_split[0] + ' => ' + t + ' => ' + hop_dst_split[1] + '\n')

    output_file_dst_one_as.close()
    output_file_filter.close()

################################################################################

if __name__ == '__main__':
	sys.exit(main())
