#!/usr/bin/env python

################################################################################
# apply_IE_AS_temporal_filter                                                  #
#                                                                              #
# Author Yves Vanaubel (yves.vanaubel@ulg.ac.be)                               #
# Date 09/01/15                                                                #
# Description: Apply a temporal filter on each target cycle.                   #
#              An LSP is valid only if it appears at least once in one of the  #
#              X cycles after the target cycle. The whole LSP is compared      #
#              (IPs and MPLS Labels)                                           #
################################################################################

# IMPORTS
import sys
import os
import re
import numpy as np
import subprocess
from collections import defaultdict
from optparse import OptionParser

def main():
	usage = "usage: %prog <output_dir> <reference_top_as_dir> <tunnels_dir> <group_ts> <output_stats_dir> <write_steps>"
	
	parser = OptionParser(usage=usage)
	
	(options, args) = parser.parse_args()
    
	#Check arg number
	if not len(args) == 6:
	   parser.print_help()
	   sys.exit(1)
	run(args)

################################################################################

def run(args):
    
    output_dir = args[0]
    reference_top_as_dir = args[1]
    tunnels_dir = args[2]
    group_ts = args[3]
    output_stats_dir = args[4]
    write_steps = args[5]
    
    # Create the output directory
    os.system('rm -rf ' + output_dir)
    os.system('mkdir ' + output_dir)
    
    # Get each cycle in chronological order
    cycles_string = subprocess.check_output('cat ' + group_ts + ' | sort -n -k 1.7,1.10 -k 1.4,1.5', shell=True)
    cycles = cycles_string.split()
    
    # The first cycle contains the AS repartition
    reference_cycle = cycles[0]
    cycles.remove(reference_cycle)
    nb_cycles = len(cycles)
    
    # Get the different AS files
    as_files_string = subprocess.check_output('ls ' + reference_top_as_dir + ' | grep -v top_ases.txt', shell=True)
    as_files = as_files_string.split()
        
    as2tunnel = defaultdict(lambda: defaultdict(list))
    as2nb_tunnels = defaultdict()
    for as_file in as_files:
        # Extract the ASn
        asn = re.search('as_(\d+).txt', as_file).group(1)
        
        # Get each tunnel
        tunnels_file = open(reference_top_as_dir + '/' + as_file)
        for tunnel in tunnels_file:
            tunnel = tunnel.strip()
            # Skip empty lines
            if tunnel == "":
                continue
        
            # Get the destination and the tunnel
            tun_dst = tunnel.split(" => ")
            tunnel_content = tun_dst[1]
            hop_dst =  tun_dst[0] + " => " +  tun_dst[2]
            
            as2tunnel[asn][tunnel_content].append(hop_dst)
            if asn in as2nb_tunnels:
                as2nb_tunnels[asn] += 1
            else:
                as2nb_tunnels[asn] = 1

    print " - Reference Cycle Read!"

    # Tunnels appearing in other cycles
    compliant_tunnels = defaultdict(lambda: defaultdict(list))
    
    # Get the tunnels of the other cycles
    considerred_cycles = 0
    for cycle in cycles:
        cycle_year = cycle.split("-")[2]
        cycle_month = cycle.split("-")[1]
        cycle_dir = tunnels_dir + '/' + cycle_year + '/' + cycle_month + '/' + cycle
        
        # Check is empty directory
        if os.listdir(cycle_dir) == []:
            continue
    
        # Decompress files if needed
        remove_tunnels = False
        if os.path.exists(cycle_dir + '/etun_dst_all.txt.gz'):
            os.system('gunzip -c ' + cycle_dir + '/etun_dst_all.txt.gz > ' + cycle_dir + '/etun_dst_all.txt' )
            remove_tunnels = True
        else:
            # No tunnels found
            if not os.path.exists(cycle_dir + '/etun_dst_all.txt'):
                continue

        considerred_cycles += 1
        
        # Get each tunnel
        tunnels_file = open(cycle_dir + '/etun_dst_all.txt')
        for tunnel in tunnels_file:
            tunnel = tunnel.strip()
            # Skip empty lines
            if tunnel == "":
                continue
    
            # Get the destination and the tunnel
            tun_dst = tunnel.split(" => ")
            tunnel_content =  tun_dst[1]
            hop_dst = tun_dst[0] + " => " + tun_dst[2]

            # Get the ASn
            i = 0
            asn_FH_LH = set()
            nodes = tunnel_content.split()
            last_index = len(nodes) - 1
            for n in nodes:
                # Get ASn between FH and LH
                if i > 0 and i < last_index:
                    if n == '*':
                        continue
                    node_data = n.split(':')
                    asn_FH_LH.add(node_data[1])
                i += 1
    
            # Check the tunnel if only in one AS
            if (not asn_FH_LH) or (len(asn_FH_LH) != 1):
                continue
            asn = asn_FH_LH.pop()

            # Check if the tunnel appears in reference cycle
            if (asn in compliant_tunnels) and (tunnel_content in compliant_tunnels[asn]):
                continue
            else:
                if (asn in as2tunnel) and (tunnel_content in as2tunnel[asn]):
                    compliant_tunnels[asn][tunnel_content] = as2tunnel[asn][tunnel_content][:]
                    del(as2tunnel[asn][tunnel_content])
        
        if (write_steps == '1') or (considerred_cycles == nb_cycles):
            output_stats_path = output_stats_dir
            output_dir_path = output_dir
            if write_steps == '1':
                os.system('mkdir ' + output_dir + '/' + str(considerred_cycles))
                os.system('mkdir ' + output_stats_dir + '/' + str(considerred_cycles))
                output_stats_path = output_stats_dir + '/' + str(considerred_cycles)
                output_dir_path = output_dir + '/' + str(considerred_cycles)
            
            output_stats_file = open(output_stats_path + '/temporal_stats' + '.txt', 'w')
            output_stats_file.write('{0:<16} {1:<16} {2:<16} {3:<16} {4:<16} {5:<16}\n'.format('#asn', '#nb_VPNs_OK', '#nb_tun_OK', '#nb_side_eff', '#nb_tunnels', '#prop_tun_OK'))
            
            # The tunnel must be writen for each destination
            
            for asn in as2nb_tunnels:
                # Highly dynamic AS
                if asn not in compliant_tunnels:
                    nb_VPNs = 0
                    output_as_file_name = output_dir_path + '/filtered_dynamic_as_' + asn + '.txt'
                    output_as_file = open(output_as_file_name, 'w')
                    nb_reinjected_LSPs = 0
                    for tunnel in as2tunnel[asn]:
                        for nd in as2tunnel[asn][tunnel]:
                            if "|" in tunnel:
                                nb_VPNs += 1
                            nd_split = nd.split(" => ")
                            nb_reinjected_LSPs += 1
                            output_as_file.write(nd_split[0] + ' => ' + tunnel + ' => ' + nd_split[1] + '\n')
                    output_stats_file.write('{0:<16} {1:<16} {2:<16} {3:<16} {4:<16} {5:<16.4f}\n'.format(str(asn), str(nb_VPNs), str(nb_reinjected_LSPs), '0', str(as2nb_tunnels[asn]), 1))
                    output_as_file.close()
                    continue
            
                # Other ASes
                ie2tunnels = defaultdict(list)
                ie2dst_as = defaultdict(set)
                ie2lsp = defaultdict(set)
                for tunnel in compliant_tunnels[asn]:
                    nodes = tunnel.split()
                    for nd in compliant_tunnels[asn][tunnel]:
                        # Extract IE
                        nd_split = nd.split(" => ")
                        ingress = nodes[0]
                        egress = nodes[-1]
                        ie = '<' + ingress + ',' + egress + '>'
                        ie2tunnels[ie].append(nd_split[0] + ' => ' + tunnel + ' => ' + nd_split[1])
                        ie2dst_as[ie].add(nd_split[1].split(":")[1])
                        ie2lsp[ie].add(tunnel)
            
                output_as_file_name = output_dir_path + '/filtered_as_' + asn + '.txt'
                output_as_file = open(output_as_file_name, 'w')

                # Delete the side effects of previous filters (overestimation of class 1)
                nb_VPNs = 0
                nb_side_effect = 0
                nb_compliant_tunnels = 0
                for ie in ie2tunnels:
                    # It remains at least 2 LSPs for this tunnel
                    if len(ie2tunnels[ie]) > 1:
                        # Need to check again static filter on IE pair
                        nb_LSPs = len(ie2lsp[ie])
                        nb_dst_as = len(ie2dst_as[ie])
                        if nb_LSPs > 1 or nb_dst_as > 1:
                            for t in ie2tunnels[ie]:
                                if "|" in t:
                                    nb_VPNs += 1
                                nb_compliant_tunnels += 1
                                output_as_file.write(t + '\n')
                        else:
                            nb_side_effect += nb_LSPs
                    else:
                        nb_side_effect += 1
                nb_tunnels = as2nb_tunnels[asn]
                output_stats_file.write('{0:<16} {1:<16} {2:<16} {3:<16} {4:<16} {5:<16.4f}\n'.format(str(asn), str(nb_VPNs), str(nb_compliant_tunnels), str(nb_side_effect), str(nb_tunnels), nb_compliant_tunnels / float(nb_tunnels)))
                output_as_file.close()
            output_stats_file.close()
        print " - Cycle " + cycle + " Done!"
        if remove_tunnels:
            os.system('rm ' + cycle_dir + '/etun_dst_all.txt' )
    print " - Temporal analyze done!"

if __name__ == '__main__':
	sys.exit(main())
