#!/usr/bin/env python

################################################################################
# get_traffic_eng_temporal_analyze.py                                          #
#                                                                              #
# Author Yves Vanaubel (yves.vanaubel@ulg.ac.be)                               #
# Date 18/09/14                                                                #
# Description: Analyze the ASes to determine how the use MPLS                  #
#              The output file as_array.txt (1 per cycle) is in the form       #
#              <Nb IPs> <Nb MPLS Expl IPs> <$2/$1> <Nb Expl Tunnels FH-LH>     #
#              Output file:                                                    #
#                  as_tunnels.txt : <asn> <tunnel> => <dst>                    #
#                  The file is sorted from AS with the highest number of       #
#                  MPLS IPs to AS having the least. It contains the tunnels    #
#                  located inside the given AS                                 #
################################################################################

# IMPORTS
import sys
import os
import re
import numpy as np
import subprocess
from collections import defaultdict
from optparse import OptionParser

def main():
	usage = "usage: %prog <output_dir> <cycle_tunnels_dir> <prefix2as_file>"
	
	parser = OptionParser(usage=usage)
	
	(options, args) = parser.parse_args()
    
	#Check arg number
	if not len(args) == 3:
	   parser.print_help()
	   sys.exit(1)
	run(args)

################################################################################

# Read the ip2as file and get the number of IPs for each AS
def read_ip2as_file(ip2as_file):
    as2nb_ips = dict()

    for line in ip2as_file:
        line.strip()
        if line == "":
            continue

        ip, asn = line.split()
        if asn in as2nb_ips:
            as2nb_ips[asn] += 1
        else:
            as2nb_ips[asn] = 1

    return as2nb_ips

################################################################################

def run(args):
    
    output_dir = args[0]
    tunnels_dir = args[1]
    prefix2as_file = args[2]
    
    os.system('rm -rf ' + output_dir)
    os.system('mkdir ' + output_dir)

    # Get the script directory
    script_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
    root_script_dir = script_dir + "/../"
    
    # Get the IPs and their ASn
    os.system('grep -v "\*" ' + tunnels_dir + '/all.ips > ' + output_dir + '/tmp_ips.txt')
    os.system('python ' + root_script_dir + '/ip2as/map_ips2asn.py ' + prefix2as_file + ' ' + output_dir + '/tmp_ips.txt ' + output_dir + '/ip2as.txt')

    # Read the IP to AS file and get the number of IPs for each AS
    ip2as_file = open(output_dir + '/ip2as.txt')
    as2nb_ips = read_ip2as_file(ip2as_file)
    as2mpls_ips = defaultdict(set)
    as2FH_LH = defaultdict(list)
    as2ie = defaultdict(set)
    
    tunnels_file = open(tunnels_dir + '/etun_dst_all.txt')
            
    # Get each tunnel in the file
    for tunnel in tunnels_file:
        tunnel = tunnel.strip()
        # Skip empty lines
        if tunnel == "":
            continue
            
        # Get the destination and the tunnel
        tun_dst = tunnel.split(" => ")
        tunnel_nodes = tun_dst[1]
        
        # Get the tunnel (FH->LH) and the IPs
        nodes = tunnel_nodes.split()
        i = 0
        tunnel_FH_LH = list()
        asn_FH_LH = set()
        last_index = len(nodes) - 1
        
        # Delete if ingress of egress unknown
        if nodes[0] == '*' or nodes[last_index] == '*':
            continue
        
        ie = '<' + nodes[0] + ',' + nodes[last_index] + '>'
        for n in nodes:
            node_data = n.split(':')
            # Between FH and LH
            if i > 0 and i < last_index:
                if n == '*':
                    tunnel_FH_LH.append(n)
                    continue
                
                # Get ASn
                asn_FH_LH.add(node_data[1])
                
                # IP:ASn:Label => MPLS tagged
                if (len(node_data) == 3):
                    as2mpls_ips[node_data[1]].add(node_data[0])
                    # Get IP and label
                    tunnel_FH_LH.append(node_data[0] + ':' + node_data[2])
                # IP:ASn => Not tagged
                else:
                    tunnel_FH_LH.append(n)
            # Ingress or Egress
            else:
                # IP:ASn:Label => MPLS tagged
                if (len(node_data) == 3):
                    as2mpls_ips[node_data[1]].add(node_data[0])
            i += 1
        
        # Add the tunnel if possible
        if asn_FH_LH:
            nb_ases = len(asn_FH_LH)
            if nb_ases == 1:
                unique_asn = asn_FH_LH.pop()
                as2FH_LH[unique_asn].append(' '.join(tunnel_FH_LH))
                as2ie[unique_asn].add(ie)

    tunnels_file.close()
    os.system('rm ' + output_dir + '/tmp_ips.txt ' + output_dir + '/ip2as.txt')

    # Sort the ASn in decreasing order of MPLS IPs
    as2nb_MPLS_IPs = dict()
    for asn in as2nb_ips:
        if asn in as2mpls_ips:
            as2nb_MPLS_IPs[asn] = len(as2mpls_ips[asn])
        else:
            as2nb_MPLS_IPs[asn] = 0

    # Write the array and the tunnels
    output_file_array = open(output_dir + '/as_array.txt', 'w')
    output_file_array.write('{0:<16} {1:<16} {2:<16} {3:<16} {4:<16} {5:<16}\n'.format('#asn', '#nb_mpls_ips', '#nb_ips', '#prop_mpls_ips', '#nb_unique_ie', '#nb_FH_to_LH'))

    for asn_nb_mpls_ips in sorted(as2nb_MPLS_IPs.items(), key=lambda x:x[1], reverse=True):
        asn = asn_nb_mpls_ips[0]
        nb_mpls_ips = asn_nb_mpls_ips[1]
        nb_ips = as2nb_ips[asn]
        prop_mpls_ips = nb_mpls_ips / float(nb_ips)
        nb_FH_LH = len(as2FH_LH[asn])
        nb_ie = len(as2ie[asn])

        # Print the array
        output_file_array.write('{0:<16} {1:<16} {2:<16} {3:<16.4f} {4:<16} {5:<16}\n'.format(asn, str(nb_mpls_ips), str(nb_ips), prop_mpls_ips, str(nb_ie), str(nb_FH_LH)))

    output_file_array.close()

if __name__ == '__main__':
	sys.exit(main())
