 #!/bin/bash

################################################################################
#   Author : Yves Vanaubel                                                     #
#   Date : 18/12/2014                                                          #
#                                                                              #
#   Run LPR on a core.                                                         #
################################################################################

shopt -s extglob

if [ $# -ne 13 ] && [ $# -ne 15 ]
then
    echo "Usage : ./run_core_LPR.sh -n <nb_group_cycles> -c <core_id> -w1 <abs_warts_store_dir> [-w2 <abs_warts_store_dir>] -t <abs_tunnels_store_dir> -l <abs_LPR_store_dir> -p <abs_prefix2as_store_dir> -warts|-tun"
    echo " Run LPR on CAIDA data."
    echo "  * nb_group_cycles : Number of cycles for each group (X, X + 1, ... X + nb_group_cycles)."
    echo "  * core_id : ID given to the core"
    echo "  * abs_warts_store_dir : Directories (absolute path) where the warts are/must be stored (w1 first, then w2 if any)"
    echo "  * abs_tunnels_store_dir : Directory (absolute path) where the tunnels are/must be stored"
    echo "  * abs_LPR_store_dir : Directory (absolute path) where LPR results must be stored"
    echo "  * abs_prefix2as_store_dir : Directory (absolute path) where prefix2as results are/must be stored"
    echo "  * -warts : If the warts are already downloaded"
    echo "  * -tun : If the tunnels are already computed"
    echo "The current directory must contains data needed for execution :"
    echo "  - scripts (must be)"
    echo "  - ts_<core_id>.txt (must be)"
    exit 1
fi

script_dir=$(dirname $0)

i=0
args=( "$@" )

nb_group_cycles=-1
core_id=0
main_warts_dir="0"
second_warts_dir="0"
tunnels_dir="0"
lpr_dir="0"
prefix2as_dir="0"

# Mode for algorithm : 2 -> warts ; 3 -> tunnels
mode=0

not_ok=0


# Check arguments
while [ $i -lt $# ]
do
    option=${args[$i]}
    let 'i = i + 1'
    arg=${args[$i]}

    if [ $option == "-c" ]
    then
        core_id=$arg
    elif [ $option == "-n" ]
    then
        nb_group_cycles=$arg
    elif [ $option == "-w1" ]
    then
        main_warts_dir=$arg
    elif [ $option == "-w2" ]
    then
        second_warts_dir=$arg
    elif [ $option == "-t" ]
    then
        tunnels_dir=$arg
    elif [ $option == "-l" ]
    then
        lpr_dir=$arg
    elif [ $option == "-p" ]
    then
        prefix2as_dir=$arg
    elif [ $option == "-warts" ]
    then
        if [ $mode -ne 0 ]
        then
            echo "-down -warts and -tun can not be used together."
            not_ok=1
        else
            mode=2
        fi
        continue
    elif [ $option == "-tun" ]
    then
        if [ $mode -ne 0 ]
        then
            echo "-down -warts and -tun can not be used together."
            not_ok=1
        else
            mode=3
        fi
        continue
    fi
    let 'i = i + 1'
done

if [ $nb_group_cycles -le 0 ]
then
    echo "The number of cycles must be higher than 0"
    not_ok=1
fi
if [ $lpr_dir == "0" ]
then
    echo "No LPR directory given"
    not_ok=1
fi
if [ $main_warts_dir == "0" ]
then
    echo "No main warts directory given"
    not_ok=1
fi
if [ $tunnels_dir == "0" ]
then
    echo "No tunnels directory given"
    not_ok=1
fi
if [ $prefix2as_dir == "0" ]
then
    echo "No prefix to AS directory given"
    not_ok=1
fi
if [ $mode -eq 0 ]
then
    echo "No mode given (down, warts, tun)"
    not_ok=1
fi

if [ ! -e $main_warts_dir ]
then
    echo "Warts directory ($main_warts_dir) does not exist"
    not_ok=1
fi
if [ $mode -eq 3 ]&&[ ! -e $tunnels_dir ]
then
    echo "Tunnels directory ($tunnels_dir) does not exist"
    not_ok=1
fi
if [ $mode -eq 3 ]&&[ ! -e $prefix2as_dir ]
then
    echo "Prefix to AS directory ($prefix2as_dir) does not exist"
    not_ok=1
fi

# Check scripts and TS file
if [ ! -e scripts ]
then
    echo "No scripts directory found in current working directory."
    not_ok=1
fi

ts_file=ts_$core_id.txt
if [ ! -e $ts_file ]
then
    echo "No TS file (ts_$core_id.txt) found in current working directory."
    not_ok=1
fi

if [ $not_ok -eq 1 ]
then
    echo "[ERROR]"
    exit 1
fi

# Create the needed directories.
if [ ! -e prefix2as ]
then
    mkdir prefix2as
fi
if [ ! -e tunnels ]
then
    mkdir tunnels
fi
if [ ! -e LPR ]
then
    mkdir LPR
fi


while read ts
do
    # Get the cycle, the corresponding month and year
    cycle_ts=$(echo $ts | awk '{print $1}')
    ts_year=$(echo $cycle_ts | awk '{split($0, d, "-"); print d[3]}')
    ts_month=$(echo $cycle_ts | awk '{split($0, d, "-"); print d[2]}')

    cycle_warts_dir=$main_warts_dir/$ts_year/$ts_month/$cycle_ts
    if [ $mode -eq 2 ]&&[ ! -e $cycle_warts_dir ]
    then
        cycle_warts_dir=$second_warts_dir/$ts_year/$ts_month/$cycle_ts
        if [ ! -e $cycle_warts_dir ]
        then
            echo "Warts for cycle $cycle_ts not found."
            continue
        fi
    fi

    #####################################
    #      STEP 1 : PREFIX2AS mapping   #
    #####################################

    # Create file architecture
    if [ ! -e prefix2as/$ts_year ]
    then
        mkdir prefix2as/$ts_year
    fi
    if [ ! -e prefix2as/$ts_year/$ts_month ]
    then
        mkdir prefix2as/$ts_year/$ts_month
    fi

    cycle_prefix2as_dir=prefix2as/$ts_year/$ts_month/$cycle_ts
    if [ $mode -eq 2 ]
    then
        # Apply prefix2AS mapping
        $script_dir/prefix2as/get_cycle_prefix2as_mapping.sh $cycle_warts_dir $cycle_prefix2as_dir 0
        if [ $? -ne 0 ]
        then
            echo "[ERROR]"
            exit 1
        fi
    else
        # Copy data
        cp -r $prefix2as_dir/$ts_year/$ts_month/$cycle_ts $cycle_prefix2as_dir
    fi

    #####################################
    #     STEP 2 : Find the tunnels     #
    #####################################

    # Create file architecture
    if [ ! -e tunnels/$ts_year ]
    then
        mkdir tunnels/$ts_year
    fi
    if [ ! -e tunnels/$ts_year/$ts_month ]
    then
        mkdir tunnels/$ts_year/$ts_month
    fi

    cycle_tunnels_dir=tunnels/$ts_year/$ts_month/$cycle_ts
    if [ $mode -eq 2 ]
    then
        # Identify tunnels
        $script_dir/tunnels/get_caida_MPLS_tunnels.sh $cycle_warts_dir $cycle_prefix2as_dir/prefix2as.txt $cycle_tunnels_dir 0
        if [ $? -ne 0 ]
        then
            echo "[ERROR]"
            exit 1
        fi
    else
        # Copy data
        cp -r $tunnels_dir/$ts_year/$ts_month/$cycle_ts $cycle_tunnels_dir
    fi
done < $ts_file

#####################################
#       STEP 3 : Apply LPR          #
#####################################

# Get the cycle, the corresponding month and year
cycle_ts=$(head -n 1 $ts_file | awk '{print $1}')
ts_year=$(echo $cycle_ts | awk '{split($0, d, "-"); print d[3]}')
ts_month=$(echo $cycle_ts | awk '{split($0, d, "-"); print d[2]}')

# Create file architecture
if [ ! -e LPR/$ts_year ]
then
    mkdir LPR/$ts_year
fi
if [ ! -e LPR/$ts_year/$ts_month ]
then
    mkdir LPR/$ts_year/$ts_month
fi
cycle_dir=LPR/$ts_year/$ts_month/$cycle_ts
if [ -e $cycle_dir ]
then
    rm -r $cycle_dir
fi
mkdir $cycle_dir

$script_dir/LPR/apply_LPRA.sh $ts_file $cycle_dir 1
if [ $? -ne 0 ]
then
    echo "[ERROR]"
    exit 1
fi

# Compressing the results
if [ $mode -eq 2 ]
then
    while read ts
    do
        # Get the cycle, the corresponding month and year
        cycle_ts=$(echo $ts | awk '{print $1}')
        ts_year=$(echo $cycle_ts | awk '{split($0, d, "-"); print d[3]}')
        ts_month=$(echo $cycle_ts | awk '{split($0, d, "-"); print d[2]}')

        # Tunnels
        cd tunnels/$ts_year/$ts_month/$cycle_ts
        gzip *
        cd - > /dev/null 2>&1

        # Prefix to AS
        cd prefix2as/$ts_year/$ts_month/$cycle_ts
        gzip *
        cd - > /dev/null 2>&1
    done < $ts_file
fi

echo "Core $core_id ended its work correctly on $(date)"
exit 0
