#!/bin/bash

################################################################################
#   Author : Yves Vanaubel                                                     #
#   Date : 16/12/2014                                                          #
#                                                                              #
#   Run LPR on warts/tunnels.                                                  #
################################################################################

shopt -s extglob

if [ $# -ne 13 ] && [ $# -ne 15 ]
then
    echo "Usage : ./run_multicore_LPR.sh -i <abs_TS_file> -n <nb_group_cycles> -w1 <abs_warts_store_dir> [-w2 <abs_warts_store_dir>] -t <abs_tunnels_store_dir> -l <abs_LPR_store_dir> -p <abs_prefix2as_store_dir> -warts|-tun"
    echo " Run LPR"
    echo "  * abs_TS_file : File (absolute path) containing the timestamps"
    echo "  * nb_group_cycles : Number of cycles for each group (X, X + 1, ... X + nb_group_cycles - 1)."
    echo "  * abs_warts_store_dir : Directories (absolute path) where the warts are/must be stored (w1 first, then w2 if any)"
    echo "  * abs_tunnels_store_dir : Directory (absolute path) where the tunnels are/must be stored"
    echo "  * abs_LPR_store_dir : Directory (absolute path) where LPR results must be stored"
    echo "  * abs_prefix2as_store_dir : Directory (absolute path) where prefix2as results are/must be stored"
    echo "  * -warts : If the warts are already downloaded"
    echo "  * -tun : If the tunnels are already computed"
    exit 1
fi

echo "Starting on $(date)"

i=0
args=( "$@" )

ts_file="0"
nb_group_cycles=0
main_warts_dir="0"
second_warts_dir="0"
tunnels_dir="0"
lpr_dir="0"
prefix2as_dir="0"

# Mode for algorithm : 2 -> warts ; 3 -> tunnels
mode=0
mode_string="0"

not_ok=0


# Check arguments
while [ $i -lt $# ]
do
    option=${args[$i]}
    let 'i = i + 1'
    arg=${args[$i]}

    if [ $option == "-i" ]
    then
        ts_file=$arg
    elif [ $option == "-n" ]
    then
        nb_group_cycles=$arg
    elif [ $option == "-w1" ]
    then
        main_warts_dir=$arg
    elif [ $option == "-w2" ]
    then
        second_warts_dir=$arg
    elif [ $option == "-t" ]
    then
        tunnels_dir=$arg
    elif [ $option == "-l" ]
    then
        lpr_dir=$arg
    elif [ $option == "-p" ]
    then
        prefix2as_dir=$arg
    elif [ $option == "-warts" ]
    then
        if [ $mode -ne 0 ]
        then
            echo "-tun -warts can not be used together."
            not_ok=1
        else
            mode=2
            mode_string=$option
        fi
        continue
    elif [ $option == "-tun" ]
    then
        if [ $mode -ne 0 ]
        then
            echo "-tun -warts can not be used together."
            not_ok=1
        else
            mode=3
            mode_string=$option
        fi
        continue
    fi
    let 'i = i + 1'
done

if [ $nb_group_cycles -le 0 ]
then
    echo "The number of cycles must be higher than 0"
    not_ok=1
fi
if [ $lpr_dir == "0" ]
then
    echo "No LPR directory given"
    not_ok=1
fi
if [ $ts_file == "0" ]
then
    echo "No timestamps file given"
    not_ok=1
fi
if [ $main_warts_dir == "0" ]
then
    echo "No main warts directory given"
    not_ok=1
fi
if [ $tunnels_dir == "0" ]
then
    echo "No tunnels directory given"
    not_ok=1
fi
if [ $prefix2as_dir == "0" ]
then
    echo "No prefix to AS directory given"
    not_ok=1
fi
if [ $mode -eq 0 ]
then
    echo "No mode given (warts, tun)"
    not_ok=1
fi

if [ $mode -eq 2 ]&&[ ! -e $main_warts_dir ]
then
    echo "Warts directory ($main_warts_dir) does not exist"
    not_ok=1
fi
if [ $mode -eq 3 ]&&[ ! -e $tunnels_dir ]
then
    echo "Tunnels directory ($tunnels_dir) does not exist"
    not_ok=1
fi
if [ $mode -eq 3 ]&&[ ! -e $prefix2as_dir ]
then
    echo "Prefix to AS directory ($prefix2as_dir) does not exist"
    not_ok=1
fi

if [ $not_ok -eq 1 ]
then
    exit 1
fi

# Get the number of processors that can work on the computer
nb_cores=1
which nproc > /dev/null 2>&1
if [ $? -eq 1 ]
then
    which sysctl > /dev/null 2>&1
    if [ $? -eq 0 ]
    then
        nb_cores=$(sysctl -n hw.ncpu)
    fi
else
    nb_cores=$(nproc)
fi

nb_computing_cores=1
if [ $nb_cores -ge 4 ]
then
    let 'nb_computing_cores = nb_cores - 2'
fi


# Create a working directory
if [ -e working ]
then
    rm -r working
fi
mkdir working

# Create a state directory
if [ -e multicore_state ]
then
    rm -r multicore_state
fi
mkdir multicore_state

# Create storage directories if needed
if [ ! -e $main_warts_dir ]
then
    mkdir $main_warts_dir
fi
if [ $second_warts_dir != "0" ]&&[ ! -e $second_warts_dir ]
then
    mkdir $second_warts_dir
fi
if [ ! -e $lpr_dir ]
then
    mkdir $lpr_dir
fi
if [ ! -e $tunnels_dir ]
then
    mkdir $tunnels_dir
fi
if [ ! -e $prefix2as_dir ]
then
    mkdir $prefix2as_dir
fi


############################################
#         STEP 1 : Prepare each core       #
############################################

core_id=1
while [ $core_id -le $nb_computing_cores ]
do
    # Copy scripts (avoid to modify each script for parallel execution.
    # Each processor will execute its scripts in its working directory (sandbox).
    core_working_dir=working/core_$core_id
    mkdir $core_working_dir
    mkdir $core_working_dir/scripts
    cp -r !(sc_warts2text|working|run_caida_LPR.sh) $core_working_dir/scripts
    let 'core_id = core_id + 1'
done

# While there are timestamps in the output file
cp $ts_file working/remaining_ts.txt
while [ -s working/remaining_ts.txt ]
do
    # Get as much timestamps as all the core need
    let 'nb_process_lines = nb_computing_cores * nb_group_cycles'
    head -n $nb_process_lines working/remaining_ts.txt > working/current_ts.txt
    nb_head_lines=$(echo $(wc -l < working/current_ts.txt))
    nb_lines=$(echo $(wc -l < working/remaining_ts.txt))
    let 'nb_rem_lines = nb_lines - nb_head_lines'
    tail -n $nb_rem_lines working/remaining_ts.txt > working/tmp.txt
    mv working/tmp.txt working/remaining_ts.txt

    ############################################
    #   STEP 2 : Distribute TS to each core    #
    ############################################

    core_id=1
    line_nb=1
    while read line
    do
        echo $line >> working/core_$core_id/ts_$core_id.txt
        if [ $line_nb -eq $nb_group_cycles ]
        then
            line_nb=1
            let 'core_id = core_id + 1'
        else
            let 'line_nb = line_nb + 1'
        fi
    done < working/current_ts.txt

    #####################################
    #      STEP 3 : Run each core       #
    #####################################

    current_core_id=1
    core_mode=$mode_string
    if [ $mode_string == "-down" ]
    then
        core_mode="-warts"
    fi
    while [ $current_core_id -lt $core_id ]
    do
        echo "Running core $current_core_id"
        cd working/core_$current_core_id
        ./scripts/run_core_LPR.sh -n $nb_group_cycles -c $current_core_id -w1 $main_warts_dir -w2 $second_warts_dir -t $tunnels_dir -l $lpr_dir -p $prefix2as_dir $core_mode > state.txt 2>&1& pid=$!
        PID_LIST+=" $pid";
        cd - > /dev/null 2>&1
        echo "Core $current_core_id run correctly"
        let 'current_core_id = current_core_id + 1'
    done
    trap "kill $PID_LIST" SIGINT
    echo "Waiting the different cores..."
    wait
    echo "All cores have their job done"

    current_core_id=1
    while [ $current_core_id -lt $core_id ]
    do
        echo -e "New Step :\nTimestamps : \n" >> multicore_state/state_core_$current_core_id.txt
        cat working/core_$current_core_id/ts_$current_core_id.txt >> multicore_state/state_core_$current_core_id.txt
        echo -e "\n" >> multicore_state/state_core_$current_core_id.txt
        cat working/core_$current_core_id/state.txt >> multicore_state/state_core_$current_core_id.txt
        echo -e "\n\n\n" >> multicore_state/state_core_$current_core_id.txt
        let 'current_core_id = current_core_id + 1'
    done

##############################################
#    STEP 4 : Transfer results for storage   #
##############################################

# NB: The tranfers are not realized by cores to avoid concurrent access to data

    echo "Transferring results for storage"
    current_core_id=1
    while [ $current_core_id -lt $core_id ]
    do
        # Get the timestamp of main cycle for storing architecture
        ts_file=working/core_$current_core_id/ts_$current_core_id.txt

        current_working_dir=working/core_$current_core_id/
        # Check for errors
        grep "ERROR" $current_working_dir/state.txt > $current_working_dir/tmp_state.txt
        if [ -s $current_working_dir/tmp_state.txt ]
        then
            cat $current_working_dir/state.txt
            exit 1
        fi
        rm $current_working_dir/state.txt $current_working_dir/tmp_state.txt

        # LPR results
        storing_ts=$(head -n 1 $ts_file | awk '{print $1}')
        storing_year=$(echo $storing_ts | awk '{split($0, d, "-"); print d[3]}')
        storing_month=$(echo $storing_ts | awk '{split($0, d, "-"); print d[2]}')
        storing_lpr_dir=$lpr_dir/$storing_year/$storing_month/$storing_ts
        if [ -e $storing_lpr_dir ]
        then
            rm -r $storing_lpr_dir
        fi
        if [ ! -e $lpr_dir/$storing_year ]
        then
            mkdir $lpr_dir/$storing_year
        fi
        if [ ! -e $lpr_dir/$storing_year/$storing_month ]
        then
            mkdir $lpr_dir/$storing_year/$storing_month
        fi
        cp -r $current_working_dir/LPR/$storing_year/$storing_month/$storing_ts $storing_lpr_dir/
        rm -r $current_working_dir/LPR


        # Tunnels and prefixes to AS results if needed
        if [ $mode -ne 3 ]
        then
            while read line
            do
                storing_ts=$(echo $line | awk '{print $1}')
                storing_year=$(echo $storing_ts | awk '{split($0, d, "-"); print d[3]}')
                storing_month=$(echo $storing_ts | awk '{split($0, d, "-"); print d[2]}')

                storing_tunnels_dir=$tunnels_dir/$storing_year/$storing_month/$storing_ts
                if [ -e $storing_tunnels_dir ]
                then
                    rm -r $storing_tunnels_dir
                fi
                if [ ! -e $tunnels_dir/$storing_year ]
                then
                    mkdir $tunnels_dir/$storing_year
                fi
                if [ ! -e $tunnels_dir/$storing_year/$storing_month ]
                then
                    mkdir $tunnels_dir/$storing_year/$storing_month
                fi
                cp -r $current_working_dir/tunnels/$storing_year/$storing_month/$storing_ts $storing_tunnels_dir/

                storing_pefix2as_dir=$prefix2as_dir/$storing_year/$storing_month/$storing_ts
                if [ -e $storing_pefix2as_dir ]
                then
                    rm -r $storing_pefix2as_dir
                fi
                if [ ! -e $prefix2as_dir/$storing_year ]
                then
                    mkdir $prefix2as_dir/$storing_year
                fi
                if [ ! -e $prefix2as_dir/$storing_year/$storing_month ]
                then
                    mkdir $prefix2as_dir/$storing_year/$storing_month
                fi
                cp -r $current_working_dir/prefix2as/$storing_year/$storing_month/$storing_ts $storing_pefix2as_dir/
            done < $ts_file
        fi
        rm -r $ts_file $current_working_dir/tunnels $current_working_dir/prefix2as

        echo "Core $current_core_id done and data for cycle $storing_ts stored"
        let 'current_core_id = current_core_id + 1'
    done
done

echo "End on $(date)"
rm -r working

exit 0
