#!/usr/bin/perl

################################################################################
#   Author : JJ. Pansiot                                                       #
#   Modified by Yves Vanaubel on 27/05/2014                                    #
#                                                                              #
#   Extract Explicit tunnels with MPLS labels from the traces obtained with    #
#   sc_warts2text                                                              #
#   Extract also the traces and output them in the format: S N1 N2 ... Ni D    #
#   Extract also the IPs.                                                      #
################################################################################


# Include the modules.
use File::Find;
use File::Basename;
use POSIX qw(strftime);
use strict;

# Some global values (T = true and F = false).
my($ADD_ANOMALY) = "F";
my($ADD_TARGET) = "T";
my($ADD_HOP) = "T";
my($PRINT_TRACES) = "F";          # SET TO FALSE IF SPACE NEEDED
my($PRINT_OPAQUES) = "F";
my($FIRST_TRACE_HOP) = "F";

my($LAMBDA_THRESH) = 3;           # Threshold at which a hop > lambda followed by -2 (hop - 1) / 2 generates a signature.
my($INVISIBLE_THRESH) = 2;



# If true (T), when a tunnel starts with a flag T2 (or T3, ...)
# and the previous router is anonymous, then, this router is considered
# as the beginning of the tunnel.
my($FIRST_MPLS_ANONYMOUS) = "F";


# Check the arguments.
if ($#ARGV < 1)
{
	print "Usage: analyze_traces_temporal.pl <traces_file> <prefix2as_file> [<output_directory>] \n";
    print " Generate 3 files named:\n";
    print "  - vp_name.etun : contains explicit tunnels\n";
    print "  - vp_name.traces : contains traces\n";
    print "  - vp_name.ips : contains IPs\n";
    print " Append the number of traces for this VP to the file vp_name.ntraces\n";
    exit;
}

# Gets the arguments.
my($inputFile) = $ARGV[0];
my($prefix2asFile) = $ARGV[1];

# Get the output directory if any and create the name of the output files, without any extensions.
my($inputFileName) = basename($inputFile);
my($outputName) = "";
if ($#ARGV == 2)
{
    my($outputDirectory) = $ARGV[2];
    $outputName = $outputDirectory . "/" . $inputFileName;
}
else
{
    $outputName = $inputFile;
}


# Open the different files that will contain the results.
open(my $explTunnels_file, ">", "$outputName.etun") || die ("Error when opening the explicit tunnels file.\n");                         # Explicit tunnels
my($traces_file);
my($opaques_file);
if ($PRINT_TRACES eq "T")
{
    open($traces_file, ">", "$outputName.traces") || die ("Error when opening the traces file.\n");                                      # Traces
}
if ($PRINT_OPAQUES eq "T")
{
    open($opaques_file, ">", "$outputName.otun") || die ("Error when opening the opaque tunnels file.\n");
}

open(my $ips_file, ">", "$outputName.ips") || die ("Error when opening the ip's file.\n");
open(my $nb_traces_file, ">>", "$outputName.ntraces") || die ("Error when opening the nb_traces file.\n");

# Array containing different information about each hop of a trace :
#   - TARGET        : Destination of the traceroute
#   - HOP           : Number of the hop
#   - IP            : IP address of the hop
#   - FLAGS         : Flags different from the type T
#   - FLAGT         : Flags of the type T
#   - RTT           : Round Trip Time for this hop during the traceroute
#   - HOP_ICMP      : Number of hops for the time exceeded ICMP response
#   - LABEL_FOND    : MPLS label for the hop
#   - LABEL_TOP     : Second MPLS label (if any)
#   - LABELS        : All labels (format L1|L2|...|Ln)
#   - FDIST         : Distance from the begining of the tunnel (begining = 1, not in tunnel = -1)
#   - RDIST         : Distance from the end of the tunnel (end = 0, not in tunnel = -1)
#   - LOCATION      : Location of the hop in the tunnel : First, Last, Internal, Unique, After, Other
#   - LAMBDA        : The lambda value in comparison with the previous hop, or 0 if it is the first.
#   - SIG           : -1 if no signature currently, otherwise the number of nodes that must still be checked. It is initialized to (hop - 1) / 2 and decreases each -2
#   - END           : True if IP = TARGET or if it is the first '*' after the last IP.
my (@trace) = ();

my(%privateIps) = ();            # Private IP addresses

# Information about the different IP's of a trace :
#   - PRESENT       : T if exists, F otherwise
#   - TARGET        : T if target of the trace
#   - LABEL         : T if the IP address is associated to 1 label
#   - TWO_LABELS    : T if the IP address is associated to one of at least two following labels 
#   - QTTL          : T if Quoted TTL
#   - UTURN         : T if U-turn value > THRESHOLD
#   - PING          : T if responded at least once to a ping
#   - TTL_PING      : Initial TTL for the ping reply
#   - TTL_ICMP      : Initial TTL for the ICMP time exceeded
my(%ipType) = ();

# Dictionary of all prefix2as mappings
my(%prefix2asn) = ();
# List of prefixes that are multilabeled
my(%multilabeled) = ();

my($line, $nextLine, $loop, $target, $hopNb, $dns, $ipAddress, $rtt, $flags, $flagt, $uturn, $hop_icmp, $fond, $top, $lsig, $hop);
my($lineNb, $fdist, $rdist, $nbHops, $nbTraces) = 0;
my($lastHop) = -1;
my($nextIp) = "";
my($no_print) = "F";


# Auxiliary functions.

# This function takes a TTL and return the number of hops that were made by the packet.
sub hopdist
{
    my($rcvdTTL) = shift;
    
    # Get the initial TTL.
    my($initTTL) = &init_ttl($rcvdTTL);
    
    return $initTTL - $rcvdTTL + 1;
}

# Get the /24 prefix of a given IP address
sub get_prefix
{
    my($ip) = shift;
    my @ip_array = split(/\./, $ip);
    return $ip_array[0] . '.' . $ip_array[1] . '.' . $ip_array[2] . '.0';
}

# This function prints the explicit tunnels in the files.
sub print_tunnel
{
    # Do not print if it was an opaque tunnel
    if ($no_print eq "T")
    {
        $no_print = "F";
        return;
    }
    
    # Last hop of the tunnel.
    my($tunnelHop) = shift;
    
    my($tunnel) = "";
    my($flagt) = "";
    my($stack) = 0;
    my($lastHopDescr) = $trace[$tunnelHop]->{"IP"} . $trace[$tunnelHop]->{"DNS"};
    my($outputSigExp) = "";
    
    # Analyzes the previous nodes of the tunnel.
    while ($trace[$tunnelHop]->{"LOCATION"} =~ /[FILU]/)
    {
        $flagt = (exists $trace[$tunnelHop]->{"FLAGT"}) ?
                 $trace[$tunnelHop]->{"FLAGT"} : "?";
        if ((exists $privateIps{$trace[$tunnelHop]->{"IP"}}) || ($trace[$tunnelHop]->{"IP"} eq "*"))
        {
            $tunnel = "* " . $tunnel;
        }
        else
        {
            my($ip_prefix) = get_prefix($trace[$tunnelHop]->{"IP"});
            my($asn) = "NA";
            if (exists $prefix2asn{$ip_prefix})
            {
                $asn = $prefix2asn{$ip_prefix};
            }
            
            if (exists $trace[$tunnelHop]->{"LABELS"})
            {
                $tunnel = $trace[$tunnelHop]->{"IP"} . "_T" . $flagt . ":" . $asn . ":" . $trace[$tunnelHop]->{"LABELS"} . " " . $tunnel;
            }
            else
            {
                $tunnel = $trace[$tunnelHop]->{"IP"} . "_T" . $flagt . ":" . $asn . " " . $tunnel;
            }
        }
        
        if (exists $trace[$tunnelHop]->{"LABEL_TOP"})
        {
            $stack++;
        }
        $tunnelHop--;
    }
    
    if ($stack > 0)
    {
        $stack = "STACK $stack " . $tunnel;
    }
    
    # Ingress router.
    if ((exists $privateIps{$trace[$tunnelHop]->{"IP"}}) || ($trace[$tunnelHop]->{"IP"} eq "*"))
    {
        $tunnel = "* " . $tunnel;
    }
    else
    {
        my($ip_prefix) = get_prefix($trace[$tunnelHop]->{"IP"});
        my($asn) = "NA";
        if (exists $prefix2asn{$ip_prefix})
        {
            $asn = $prefix2asn{$ip_prefix};
        }
        $tunnel = $trace[$tunnelHop]->{"IP"} . ":" . $asn . " " . $tunnel;
    }
    
    # Add the next hop.
    if ((exists $privateIps{$nextIp}) || ($nextIp eq "*")  || ($FIRST_TRACE_HOP eq "T"))
    {
        $tunnel = $tunnel . "*";
    }
    else
    {
        my($ip_prefix) = get_prefix($nextIp);
        my($asn) = "NA";
        if (exists $prefix2asn{$ip_prefix})
        {
            $asn = $prefix2asn{$ip_prefix};
        }
        $tunnel = $tunnel . $nextIp . ":" . $asn;
    }
    
    # Check if the tunnel is empty. TODO => empty tunnels (due to private IP addresses) => can not be identified correctly
    if (!($tunnel =~ /[0-9]/))
    {
        # print "WARNING : Empty tunnel ($lastHopDescr)\n";
        return;
    }

    # Add the target.
    if ($ADD_TARGET eq "T")
    {
        my($ip_prefix) = get_prefix($trace[$tunnelHop]->{"TARGET"});
        my($asn) = "NA";
        if (exists $prefix2asn{$ip_prefix})
        {
            $asn = $prefix2asn{$ip_prefix};
        }
        $tunnel = $tunnel . " => " . $trace[$tunnelHop]->{"TARGET"} . ":$asn";
    }
    
    # Add the hop number.
    if ($ADD_HOP eq "T")
    {
        $tunnel = $trace[$tunnelHop]->{"HOP"} . "=> " . $tunnel;
    }
    
    print $explTunnels_file "$tunnel\n";
}



# Read prefix2ASn file

open(my $prefix2as_file, "<", $prefix2asFile) || die ("Error when opening $prefix2asFile");
my($prefix2as_line);
while ($prefix2as_line = <$prefix2as_file>)
{
    chop($prefix2as_line);
    if ($prefix2as_line eq "")
    {
        next;
    }
    
    my @prefix2as_array = split(' ', $prefix2as_line);
    my($prefix) = $prefix2as_array[0];
    if (exists $prefix2asn{$prefix})
    {
        if (!exists $multilabeled{$prefix})
        {
            $multilabeled{$prefix} = $prefix2as_array[1];
        }
    }
    else
    {
        $prefix2asn{$prefix} = $prefix2as_array[1];
    }
}
close($prefix2as_file);

# Delete the multilabeled ASes
for my $key (keys %multilabeled)
{
    $prefix2asn{$key} = "NA";
}

# Read the file.

# Example of format :
# traceroute from 195.148.124.66 to 12.100.180.192
# 12  (12.122.1.2)  125.662 ms !T2 MPLS Label 0 TTL=255 MPLS Label 22927 TTL=255

# Open the input file.
open(my $input_file, "<", $inputFile) || die ("Error when opening $inputFileName");

$nextLine = <$input_file>;
my($firstline) = 0;
while (($line = $nextLine) && ($nextLine = <$input_file>))
{
    $firstline += 1;
    
    # Get the next hop IP address.
    $nextIp = $nextLine;
    chop($nextIp);
        
    # Check if there is a new trace.
    if ($nextIp =~ /^traceroute/)
    {
        $nextIp = "*";
    }
    # Check if there is an IP address between brackets.
    elsif ($nextIp =~ /\(.*\)/)
    {
        $nextIp =~ s/[^\(]*\(//;     # Delete all the characters, different from '(' before the last '('.
        $nextIp =~ s/\).*//;         # Delete all the characters after the last ')'.
        
        # Check if it is a private IP address
        if (($nextLine =~ /rfc1918$/) || ($nextLine =~ /local$/))
        {
            $privateIps{$nextIp} = "T";
        }
    }
    # No IP address
    else
    {
        $nextIp = "*";
    }
    
    
    # Analyze the current line.
    chop($line);
    
    # Check if it is a new trace.
    if ($line =~ /^traceroute/)
    {
        $FIRST_TRACE_HOP = "F";
        if (($PRINT_TRACES eq "T") && ($firstline != 1))
        {
            print $traces_file "\n";
        }
        
        # traceroute from 195.148.124.66 to 12.100.180.192
        $nbTraces++;
        if (($nbTraces > 1) && ($lineNb > 0))
        {
            # Check if the previous trace is complete (signature, tunnel).
            if ($trace[$lineNb - 1]->{"LOCATION"} eq "I")
            {
                # The tunnel is not closed. The previous hop is the last of the tunnel.
                $trace[$lineNb - 1]->{"LOCATION"} = "L" ;
                # Print the tunnel.
                &print_tunnel($lineNb - 1);
            }
            elsif($trace[$lineNb - 1]->{"LOCATION"} eq "F")
            {
                # The tunnel is not closed. The previous hop is unique in the tunnel.
                $trace[$lineNb - 1]->{"LOCATION"} eq "U";
                # Print the tunnel.
                &print_tunnel($lineNb - 1);
            }
        }
        
        # Gets the target of the trace.
        $target = $line;
        $target =~ s/.*to //;        # Delete all the characters before the last 'to '.
        
        # Re-initialize the different variables for a new trace.
        $loop = "F";
        @trace = ();
        $fdist = -1;
        $lineNb = 0 ;
        $lastHop = -1 ;
        $hopNb = 0;
    }
    # Check if there is a loop.
    elsif ($line =~ /LOOP/)
    {
        $FIRST_TRACE_HOP = "F";
        $loop = "T";
    }
    # Check if we are already in a loop.
    elsif ($loop eq "T")
    {
        $FIRST_TRACE_HOP = "F";
    }
    # Check if there is a hop of the trace.
    elsif ($line =~ /^[ ]*[0-9][0-9]* / )
    {
        $FIRST_TRACE_HOP = "F";
        # 12  (12.122.1.2)  125.662 ms !T2 MPLS Label 0 TTL=255 MPLS Label 22927 TTL=255
        $nbHops++;
        
        # Get the number of the hop.
        $hop = $line;
        $hop =~ s/^([ ]*[0-9][0-9]* ).*/\1/;
        if (($lineNb > 0) && ($hop !=  $trace[$lineNb - 1]->{"HOP"} + 1))
        {
            print "ERROR : The hops are not consecutive.\n";
        }
        $trace[$lineNb]->{"HOP"} = $hop;
        
        # Get the target of the trace.
        $trace[$lineNb]->{"TARGET"} = $target;
        
        # Deletes the number of the hop from the input.
        # (12.122.1.2)  125.662 ms !T2 MPLS Label 0 TTL=255 MPLS Label 22927 TTL=255
        $line =~ s/^[ 0-9]*[0-9][ ]+//;
        
        # Checks if there was a response.
        if ($line eq "*")
        {
            # In this case, the router is anonymous.
            $trace[$lineNb]->{"DNS"} = "*";
            $trace[$lineNb]->{"IP"} = "*";
            $trace[$lineNb]->{"LOCATION"} = "*";
            $fdist = -1;
            
            # If there is a tunnel, it is closed and printed.
            if ($trace[$lineNb - 1]->{"LOCATION"} eq "I")
            {
                $trace[$lineNb - 1]->{"WHERE"} = "L";
                $nextIp = "*";
                &print_tunnel($lineNb - 1);
            }
            elsif ($trace[$lineNb - 1]->{"LOCATION"} eq "F")
            {
                $trace[$lineNb - 1]->{"LOCATION"} = "U";
                $nextIp = "*";
                &print_tunnel($lineNb - 1);
            }
            
            $lineNb++;

            if ($PRINT_TRACES eq "T")
            {
                print $traces_file "* ";
            }
            next;
        }
        
        
        
        # The router is not anonymous.

        # Get the IP address.
        $ipAddress = $line;
        $ipAddress =~ s/^\(//;          # Delete all the characters before the last '('.
        $ipAddress =~ s/\).*//;         # Delete all the character after the last ')'.
        $lastHop = $lineNb;
        $trace[$lineNb]->{"IP"} = $ipAddress;
        $trace[$lineNb]->{"DNS"} = "empty";
        
        # Register the IP address
        if (($ipAddress ne "*") && ($trace[$lineNb]->{"TARGET"} ne $ipAddress))
        {
            if (!exists $privateIps{$ipAddress})
            {
                print $ips_file "$ipAddress \n";
            }
        }
        
        # Complete the IP type.
        $ipType{$ipAddress}->{"PRESENT"} = "T";
        if ($ipAddress eq $target)
        {
            $ipType{$ipAddress}->{"TARGET"} = "T";
        }
        
        $line =~ s/^[^\)]*\)  //;        # Delete the IP address from the input.
        # 125.662 ms !T2 MPLS Label 0 TTL=255 MPLS Label 22927 TTL=255
        
        # Get the RTT.
        $rtt = $line;
        $rtt =~ s/ .*$//;               # Delete all the characters after the first white space.
        $trace[$lineNb]->{"RTT"} = $rtt;
        
        $line =~ s/^[0-9.]+ ms +//;      # Delete the RTT.
        # !T2 MPLS Label 0 TTL=255 MPLS Label 22927 TTL=255
        
        # Check if there is flags, different from T.
        if ($line =~ /^![^T]/)
        {
            $flags = $line;
            $flags =~ s/ .*//;          # Delete all the characters after the first white space.
            $trace[$lineNb]->{"FLAGS"} = $flags;
            
            $line =~ s/^[^ ]+ //;       # Delete the flags.
        }
        
        # Check if there is a T flag.
        if ($line =~ /^!T/)
        {
            $flagt = $line;
            $flagt =~ s/^!T([0-9]+).*/\1/;     # Get the number associated to the flag.
            $trace[$lineNb]->{"FLAGT"} = $flagt;
            
            $line =~ s/^[^ ]+ //;       # Delete the flag.
        }
        
        # Get the different MPLS labels, if any.
        # MPLS Label 0 TTL=255 MPLS Label 22927 TTL=255
        if ($line =~ /MPLS/)
        {
            # The RFC4950 is respected.
            $ipType{$ipAddress}->{"LABEL"} = "T";
            if (($lineNb > 0) && (exists $trace[$lineNb - 1]->{"LABEL_FOND"}))
            {
                # There is at least 2 labels.
                $ipType{$ipAddress}->{"TWO_LABELS"} = "T";
                $ipType{$trace[$lineNb - 1]->{"IP"}}->{"TWO_LABELS"} = "T";
            }
            
            my($tag) = "T";
            if ($ipAddress ne $target)
            {
                $tag = "L";
            }
            if ($PRINT_TRACES eq "T")
            {
                if ((exists $privateIps{$ipAddress}) || ($ipAddress eq "*"))
                {
                    print $traces_file "* ";
                }
                else
                {
                    my($ip_prefix) = get_prefix($ipAddress);
                    my($asn) = "NA";
                    if (exists $prefix2asn{$ip_prefix})
                    {
                        $asn = $prefix2asn{$ip_prefix};
                    }
                    print $traces_file "$ipAddress:$asn,$tag ";
                }
            }
            
            # Gets the MPLS TTL.
            my($mplsttl) = $line;
            $mplsttl =~ s/.*TTL=//;
            $mplsttl =~ s/ .*//;
 
            # Check if it is the first node of a tunnel.
            my($is_opaque) = "F";
            if ($fdist == -1)
            {
                # Check if the previous router is anonymous.
                if (($FIRST_MPLS_ANONYMOUS eq "T") &&
                    ($lineNb > 0) &&
                    ($trace[$lineNb - 1]->{"DNS"} eq "*") &&
                    (exists $trace[$lineNb]->{"FLAGT"}) &&
                    ($trace[$lineNb]->{"FLAGT"} > 1))
                {
                    if (($lineNb > 1) && ($trace[$lineNb - 2]->{"DNS"} eq "*") &&
                        ($trace[$lineNb]->{"FLAGT"} > 2))
                    {
                        # We have flag > 2 and 2 anonymous routers.
                        $fdist = 3;
                    }
                    else
                    {
                        $fdist = 2;
                    }
                    
                    # The node is internal in the tunnel.
                    $trace[$lineNb]->{"LOCATION"} = "I";
                }
                # Check if the node is the first in the tunnel.
                else
                {
                    $fdist = 1;
                    $trace[$lineNb]->{"LOCATION"} = "F";
                }
                
                # Opaque tunnels
                if (($mplsttl < 255) && ($mplsttl > 200))
                {
                    # Specifies this tunnel must not be printed
                    $no_print = "T";
                    $is_opaque = "T";
                }
            }
            else
            {
                # We are already in the tunnel.
                $trace[$lineNb]->{"LOCATION"} = "I";
                $fdist++;
            }
            $trace[$lineNb]->{"FDIST"} = $fdist;
            
            # Get the labels
            my($labels) = "";
            my($labels_line) = $line;
            while ($labels_line =~ /MPLS/)
            {
                my($new_label) = $labels_line;
                $labels_line =~ s/^ *MPLS Label [0-9]+ TTL=[0-9]+//;
                $new_label =~ s/^ *MPLS Label ([0-9]+).*$/\1/;
                if ($labels eq "")
                {
                    $labels = $new_label;
                }
                else
                {
                    $labels = $new_label . "|" . $labels;
                }
            }
            $trace[$lineNb]->{"LABELS"} = $labels;
            
            # Check if there is several labels.
            if ($line =~ / MPLS / )
            {
                # First label.
                $top = $line;
                $top =~ s/^MPLS Label ([0-9]+).*$/\1/;
                $line =~ s/^.*MPLS/ MPLS/;
                
                $fond = $line;
                $fond =~ s/ MPLS Label ([0-9]+).*$/\1/;
 
                $trace[$lineNb]->{"LABEL_FOND"} = $fond;
                $trace[$lineNb]->{"LABEL_TOP"} = $top;
            }
            else
            {
                $fond = $line;
                $fond =~ s/MPLS Label ([0-9]+).*$/\1/;
                $trace[$lineNb]->{"LABEL_FOND"} = $fond;
            }
            
            if ($is_opaque eq "T")
            {
                # If private node => deleted
                if (exists $privateIps{$ipAddress})
                {
                    $lineNb++;
                    next;
                }
                
                my($ip_prefix) = get_prefix($ipAddress);
                my($asn) = "NA";
                if (exists $prefix2asn{$ip_prefix})
                {
                    $asn = $prefix2asn{$ip_prefix};
                }
                
                $mplsttl = 256 - $mplsttl;
                my($opaque) = $ipAddress . "_M" . $mplsttl . ":" . $asn . ":" . $trace[$lineNb]->{"LABELS"};
                
                # Add ingress
                my($ingress_addr) = $trace[$lineNb - 1]->{"IP"};
                if ((exists $privateIps{$ingress_addr}) || ($ingress_addr eq "*"))
                {
                    $opaque = "* " . $opaque;
                }
                else
                {
                    my($ip_prefix) = get_prefix($ingress_addr);
                    my($asn) = "NA";
                    if (exists $prefix2asn{$ip_prefix})
                    {
                        $asn = $prefix2asn{$ip_prefix};
                    }
                    $opaque = $ingress_addr . ":" . $asn . " " . $opaque;
                }
                
                # Add egress
                if ((exists $privateIps{$nextIp}) || ($nextIp eq "*") || ($FIRST_TRACE_HOP eq "T"))
                {
                    $opaque = $opaque . " *";
                }
                else
                {
                    my($ip_prefix) = get_prefix($nextIp);
                    my($asn) = "NA";
                    if (exists $prefix2asn{$ip_prefix})
                    {
                        $asn = $prefix2asn{$ip_prefix};
                    }
                    $opaque = $opaque . " " . $nextIp . ":" . $asn;
                }
                
                # Add target
                if ($ADD_TARGET eq "T")
                {
                    my($target_addr) = $trace[$lineNb]->{"TARGET"};
                    my($ip_prefix) = get_prefix($target_addr);
                    my($asn) = "NA";
                    if (exists $prefix2asn{$ip_prefix})
                    {
                        $asn = $prefix2asn{$ip_prefix};
                    }

                    $opaque = $opaque . " => " . $target_addr . ":" . $asn;
                }
                
                # Add hop number
                if ($ADD_HOP eq "T")
                {
                    $opaque = $trace[$lineNb - 1]->{"HOP"} . " => " . $opaque;
                }
                if ($PRINT_OPAQUES eq "T")
                {
                    print $opaques_file "$opaque\n";
                }
            }
        }
        else
        {
            my($tag) = "T";
            if ($ipAddress ne $target)
            {
                $tag = "N";
            }
            
            if ($PRINT_TRACES eq "T")
            {
                if ((exists $privateIps{$ipAddress}) || ($ipAddress eq "*"))
                {
                    print $traces_file "* ";
                }
                else
                {
                    my($ip_prefix) = get_prefix($ipAddress);
                    my($asn) = "NA";
                    if (exists $prefix2asn{$ip_prefix})
                    {
                        $asn = $prefix2asn{$ip_prefix};
                    }
                    print $traces_file "$ipAddress:$asn,$tag ";
                }
            }
            
            # No MPLS label.
            if ($fdist == 1)
            {
                # The tunnel has only 1 router.
                $trace[$lineNb - 1]->{"LOCATION"} = "U";
                $trace[$lineNb]->{"LOCATION"} = "A";
                $nextIp = $ipAddress;
                &print_tunnel($lineNb - 1);
            }
            elsif ($fdist >= 2)
            {
                # The tunnel has at least 2 hops.
                $trace[$lineNb - 1]->{"W"} = "L";
                $trace[$lineNb]->{"LOCATION"} = "A";
                $nextIp = $ipAddress;
                &print_tunnel($lineNb - 1);
            }
            else
            {
                # We are out of the tunnel.
                $trace[$lineNb]->{"LOCATION"} = "O";
            }
            
            $fdist = -1;
            $trace[$lineNb]->{"FDIST"} = $fdist;
        }
        $lineNb++;
    }
}

if ($PRINT_TRACES eq "T")
{
    print $traces_file "\n";
    close($traces_file);
}
print $nb_traces_file "$nbTraces\n";

# Close the files.
close($explTunnels_file);
` sort -u -o  $outputName.etun $outputName.etun ` ;
if ($PRINT_OPAQUES eq "T")
{
    close($opaques_file) ;
    ` sort -u -o  $outputName.otun $outputName.otun ` ;
}
close($ips_file);

` sort -u -o  $outputName.ips $outputName.ips ` ;

