#!/bin/bash

################################################################################
#   Author : Yves Vanaubel                                                     #
#   Date : 07/01/2015                                                          #
#   Derive MPLS tunnels for a cycle                                            #
#   If the extension of prefix_to_asn_file is .gz, it is decompressed          #
################################################################################

shopt -s extglob

if [ $# -ne 4 ]
then
    echo 'Usage : ./get_caida_MPLS_tunnels.sh <cycle_warts_dir> <[compressed_]prefix2asn_file> <output_dir> <memory_saving>'
    exit 1
fi

script_dir=$(dirname $0)
root_script_dir=$script_dir/../

cycle_warts_dir=$1
prefix2asn_arg=$2
output_dir=$3
memory_saving=$4

cycle_ts=$(basename $cycle_warts_dir)

if [ -e $output_dir ]
then
    rm -rf $output_dir
fi
mkdir $output_dir

cycle_results_dir=$output_dir/results
cycle_data_dir=$output_dir/raw_data
cycle_tmp_dir=$output_dir/tmp
mkdir $cycle_data_dir
mkdir $cycle_results_dir
mkdir $cycle_tmp_dir

# Unzip prefix to AS file if needed
if [ $(echo $prefix2asn_arg | awk -F . '{print $NF}') == "gz" ]
then
    unzip_prefix2asn_file=$(basename $prefix2asn_arg .gz)
    gunzip -c $prefix2asn_arg > $output_dir/unzip_prefix2asn_file
    prefix2asn_file=$output_dir/unzip_prefix2asn_file
else
    cp $prefix2asn_arg $output_dir/unzip_prefix2asn_file
    prefix2asn_file=$output_dir/unzip_prefix2asn_file
fi

# Consider each VP for the cycle
counter=0
nb_deleted_traces=0
nb_VPs=$(echo $(ls $cycle_warts_dir | wc -l))

# Analyze the warts file for each VP
for d in `ls $cycle_warts_dir`;
do
    let 'counter = counter + 1'
    echo "VP : $d - $counter / $nb_VPs"
    echo $(date)
    if test -d $cycle_warts_dir/$d
    then
        vp_data_dir=$cycle_data_dir/$d
        mkdir $vp_data_dir

        ptmp=$cycle_warts_dir/$d
        if [ -e $ptmp/warts ]
        then
            ptmp=$ptmp/warts
            let 'j = 0'
            # Analyzes the warts files.
            for w in `ls $ptmp/*.warts.gz`;
            do
                # Decompress warts file
                warts_file=$(basename $w .gz)
                gunzip -c $w > $warts_file
                echo "Reading warts..."
                sc_warts2text $warts_file > $cycle_tmp_dir/"$d"_"$j" 2> $cycle_tmp_dir/stderr.txt
                error=$?
                rm -rf $warts_file
                echo "Warts read"

                # Check if any error (Note : sc_mpls_anal_original does not write truncated data in the output file)
                grep -v "truncated" $cycle_tmp_dir/stderr.txt > $cycle_tmp_dir/stderr2.txt
                mv $cycle_tmp_dir/stderr2.txt $cycle_tmp_dir/stderr.txt
                if [ -s $cycle_tmp_dir/stderr.txt ]
                then
                    echo "Error for $d :"
                    cat $cycle_tmp_dir/stderr.txt
                    error=1
                fi
                # Delete the last traceroute if needed (may be corrupted, incomplete)
                if [ $error -ne 0 ]
                then
                    trunc=$(cat -n $cycle_tmp_dir/"$d"_"$j" | grep "traceroute" | tail -n 1 | awk '{print $1}')
                    let 'trunc = trunc - 1'
                    if [ $trunc -gt 0 ]
                    then
                        head -n $trunc $cycle_tmp_dir/"$d"_"$j" > $cycle_tmp_dir/updated
                        mv $cycle_tmp_dir/updated $cycle_tmp_dir/"$d"_"$j"
                    else
                        # The file seems completely corrupted
                        echo -n "" > $cycle_tmp_dir/"$d"_"$j"
                    fi
                fi
                let 'j = j + 1'
            done

            cd $cycle_tmp_dir/
            cat "$d"_* > $d
            rm -f !($d)
            cd - > /dev/null 2>&1

            # Clean the traces if needed
            echo "Cleaning traces..."
            tmp_nb=$(python $root_script_dir/cleaning/clean_caida_traces.py $cycle_tmp_dir/$d $cycle_tmp_dir/clean)
            let 'nb_deleted_traces = nb_deleted_traces + tmp_nb'
            mv $cycle_tmp_dir/clean $cycle_tmp_dir/$d
            echo "Cleaning done"

            # Extract the tunnels from the data.
            echo "Extracting tunnels..."
            $script_dir/extract_tunnels_from_warts_caida.pl $cycle_tmp_dir/$d $prefix2asn_file $vp_data_dir
            echo "Tunnels extracted"

            # Delete hop number and "_T..U.."
            sed -e "s/_[^:]*//g" $vp_data_dir/"$d".etun > $vp_data_dir/"$d"_clean.etun

            # Check the target is not in the tunnel
            awk '{if ((index($(NF - 2), $NF) != 1) && (index($(NF - 3), $NF) != 1)) print $0}' $vp_data_dir/"$d"_clean.etun > $vp_data_dir/etun_dst.txt

            # Concatenate for the whole campaign
            cat $vp_data_dir/etun_dst.txt >> $cycle_data_dir/etun_dst_all.txt
            cat $vp_data_dir/"$d".ntraces >> $cycle_data_dir/all.ntraces
            cat $vp_data_dir/"$d".ips >> $cycle_data_dir/all.ips

            rm -r $vp_data_dir $cycle_tmp_dir/$d
        fi
    fi
done
echo

# Sorting
echo "Processing global results for this cycle."
sort -u -o $cycle_data_dir/all.ips $cycle_data_dir/all.ips

# Number of traces containing at least 1 explicit MPLS tunnel
nb_dst_explicit_MPLS=$(echo $(awk '{print $NF}' $cycle_data_dir/etun_dst_all.txt | sort -u | wc -l))

# Number of MPLS IPs
nb_MPLS_IPs=$(awk '{for (i = 3; i <= NF; i++) {n = split($i,a,":"); if (n == 3) {print a[1]}}}' $cycle_data_dir/etun_dst_all.txt | sort -u | wc -l)

# Number of traces
nb_traces=$(awk 'BEGIN{c=0}{c+=$1}END{print c}' $cycle_data_dir/all.ntraces)
nb_ips=$(wc -l < $cycle_data_dir/all.ips)

# Stats
prop_traces_expl=$(echo "$nb_dst_explicit_MPLS $nb_traces" | awk '{printf "%.3f", $1 / $2}' | sed -e 's/,/./g')
prop_MPLS_IPs=$(echo "$nb_MPLS_IPs $nb_ips" | awk '{printf "%.3f", $1 / $2}' | sed -e 's/,/./g')

echo "Number of deleted traces for this campaign : $nb_deleted_traces" >> $cycle_results_dir/stats_labels.txt
echo "Number of (remaining) traces in this campaign : $nb_traces" >> $cycle_results_dir/stats_labels.txt
echo "Number of IP addresses for this campaign : $nb_ips" >> $cycle_results_dir/stats_labels.txt
echo "Number of (remaining) traces containing at least 1 explicit MPLS tunnel : $nb_dst_explicit_MPLS $prop_traces_expl" >> $cycle_results_dir/stats_labels.txt
echo "Number of MPLS tagged IPs : $nb_MPLS_IPs $prop_MPLS_IPs" >> $cycle_results_dir/stats_labels.txt

cd $cycle_results_dir
rm -rf !(stats_labels.txt)
cd - > /dev/null 2>&1
mv $cycle_data_dir/etun_dst_all.txt $cycle_results_dir
mv $cycle_data_dir/all.ips $cycle_results_dir

mv $cycle_results_dir/* $output_dir
rm -rf $cycle_results_dir
rm -rf $cycle_data_dir

rm -rf $prefix2asn_file $cycle_tmp_dir

if [ $memory_saving -eq 1 ]
then
    # Compress the results
    cd $output_dir
    gzip *
    cd - > /dev/null 2>&1
fi

echo "The analyze is over."

exit 0
